# Copyright 2015,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import acos_client

from a10_neutron_lbaas.db import models
import a10_neutron_lbaas.plumbing_hooks as hooks

import test_base

dev1 = {'name': 'dev1'}
dev2 = {'name': 'dev2'}
dev3 = {'name': 'dev3'}
dev4 = {'name': 'dev4'}
dev5 = {'name': 'dev5'}
dev6 = {'name': 'dev6'}

devices1 = {'dev1': dev1}
devices2 = {
    'dev1': dev1,
    'dev2': dev2,
    'dev3': dev3,
    'dev4': dev4,
    'dev5': dev5,
    'dev6': dev6,
}

TENANT_ID = 'xxx'

EXPECTED_DEV1 = acos_client.Hash(devices1.keys()).get_server(TENANT_ID)
HARDCODE_RESULT1 = 'dev1'
EXPECTED_DEV2 = acos_client.Hash(devices2.keys()).get_server(TENANT_ID)
HARDCODE_RESULT2 = 'dev2'


class TestSelectDevice(test_base.UnitTestBase):

    def test_test_setup(self):
        self.assertEqual(EXPECTED_DEV1, HARDCODE_RESULT1)
        self.assertEqual(EXPECTED_DEV2, HARDCODE_RESULT2)
        self.assertNotEqual(EXPECTED_DEV1, EXPECTED_DEV2)

    def test_select_device_hash(self):
        h = hooks.PlumbingHooks(None, devices=devices1)
        d = h._select_device_hash(TENANT_ID)
        self.assertEqual(d['name'], EXPECTED_DEV1)

        h = hooks.PlumbingHooks(None, devices=devices2)
        d = h._select_device_hash(TENANT_ID)
        self.assertEqual(d['name'], EXPECTED_DEV2)

    def test_select_device_db(self):
        h = hooks.PlumbingHooks(None, devices=devices1)
        db = self.open_session()
        d = h._select_device_db(TENANT_ID, db_session=db)
        self.assertEqual(d['name'], EXPECTED_DEV1)

        db = self.open_session()
        z = list(db.query(models.A10TenantBinding))
        self.assertEqual(z[0].tenant_id, TENANT_ID)
        self.assertEqual(z[0].device_name, EXPECTED_DEV1)

        h = hooks.PlumbingHooks(None, devices=devices2)
        db = self.open_session()
        d = h._select_device_db(TENANT_ID, db_session=db)
        self.assertEqual(d['name'], EXPECTED_DEV1)  # 1 is not a typo



# Copyright 2015,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

try:
    import neutron_lib.exceptions as lib_exceptions
except ImportError:
    lib_exceptions = None

try:
    import neutron.common.exceptions as old_exceptions
except ImportError:
    old_exceptions = None


def _find(*args):
    for f in args:
        try:
            return f()
        except AttributeError:
            pass


NotFound = _find(lambda: old_exceptions.NotFound,
                 lambda: lib_exceptions.NotFound)
InUse = _find(lambda: old_exceptions.InUse,
              lambda: lib_exceptions.InUse)
Conflict = _find(lambda: old_exceptions.Conflict,
                 lambda: lib_exceptions.Conflict)



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import mock

import a10_neutron_lbaas.a10_exceptions as a10_ex
import fake_objs
import test_base

LOG = logging.getLogger(__name__)


def return_one(*args):
    return 1


def return_two(*args):
    return 2


class TestMembers(test_base.HandlerTestBase):

    def set_count_1(self):
        self.a.member.neutron.member_count = return_one

    def set_count_2(self):
        self.a.member.neutron.member_count = return_two

    def test_get_ip(self):
        m = fake_objs.FakeMember(pool=mock.MagicMock())
        self.a.member.neutron.member_get_ip(None, m, False)
        self.print_mocks()
        self.a.neutron.member_get_ip.assert_called_with(
            None, m, False)

    def test_get_name(self):
        m = fake_objs.FakeMember(pool=mock.MagicMock())
        z = self.a.member._get_name(m, '1.1.1.1')
        self.assertEqual(z, '_get-o_1_1_1_1_neutron')

    def test_count(self):
        self.a.member.neutron.member_count(
            None, fake_objs.FakeMember(pool=mock.MagicMock()))

    def _test_create(self, admin_state_up=True, uuid_name=False, conn_limit=8000000):
        if uuid_name:
            old = self.a.config.get('member_name_use_uuid')
            self.a.config._config.member_name_use_uuid = True

        m = fake_objs.FakeMember(admin_state_up=admin_state_up,
                                 pool=mock.MagicMock())
        ip = self.a.member.neutron.member_get_ip(None, m, True)
        if uuid_name:
            name = m.id
        else:
            name = self.a.member._get_name(m, ip)
        self.a.member.create(None, m)

        if admin_state_up:
            status = self.a.last_client.slb.UP
        else:
            status = self.a.last_client.slb.DOWN
        # TODO(mdurrant) - can we do this a better way without an if?
        if conn_limit < 1 or conn_limit > 8000000:
            self.a.last_client.slb.server.create.assert_called_with(
                name, ip,
                status=status,
                config_defaults=mock.ANY,
                axapi_args={'server': {}})
        else:
            self.a.last_client.slb.server.create.assert_called_with(
                name, ip,
                status=status,
                config_defaults=mock.ANY,
                axapi_args={'server': {'conn-limit': conn_limit}})
        self.a.last_client.slb.service_group.member.create.assert_called_with(
            m.pool.id, name, m.protocol_port, status=status,
            axapi_args={'member': {}})
        if uuid_name:
            self.a.config._config.member_name_use_uuid = old

    def test_create_connlimit(self):
        for k, v in self.a.config.get_devices().items():
            v['conn-limit'] = 1337
        self._test_create(conn_limit=1337)

    def test_create_connlimit_oob(self):
        for k, v in self.a.config.get_devices().items():
            v['conn-limit'] = 8000001
        try:
            self._test_create(conn_limit=8000001)
        except a10_ex.ConnLimitOutOfBounds:
            pass

    def test_create_connlimit_uob(self):
        for k, v in self.a.config.get_devices().items():
            v['conn-limit'] = 0
        try:
            self._test_create(conn_limit=0)
        except a10_ex.ConnLimitOutOfBounds:
            pass

    def test_update_down(self):
        m = fake_objs.FakeMember(False, pool=mock.MagicMock())
        ip = self.a.member.neutron.member_get_ip(None, m, True)
        name = self.a.member._get_name(m, ip)
        self.a.member.update(None, m, m)

        self.a.last_client.slb.service_group.member.update.assert_called_with(
            m.pool.id, name, m.protocol_port, self.a.last_client.slb.DOWN,
            axapi_args={'member': {}})

    def test_delete(self):
        m = fake_objs.FakeMember(False, pool=mock.MagicMock())
        ip = self.a.member.neutron.member_get_ip(None, m, True)

        self.set_count_1()
        self.a.member.delete(None, m)

        self.a.last_client.slb.server.delete(ip)

    def test_delete_count_gt_one(self):
        m = fake_objs.FakeMember(False, pool=mock.MagicMock())
        ip = self.a.member.neutron.member_get_ip(None, m, True)
        name = self.a.member._get_name(m, ip)

        self.set_count_2()
        self.a.member.delete(None, m)

        self.a.last_client.slb.service_group.member.delete.assert_called_with(
            m.pool_id, name, m.protocol_port)

    def _test_create_expressions(self, os_name, pattern, expressions=None):
        self.a.config.get_member_expressions = self._get_expressions_mock
        expressions = expressions or self.a.config.get_member_expressions()
        expected = expressions.get(pattern, {}).get("json", {})
        admin_state = self.a.last_client.slb.UP
        m = fake_objs.FakeMember(admin_state_up=admin_state,
                                 pool=mock.MagicMock())

        m.name = os_name

        handler = self.a.member
        handler.create(None, m)

        # s = str(self.a.last_client.mock_calls)
        self.a.last_client.slb.server.create.assert_called_with(
            mock.ANY,
            mock.ANY,
            status=mock.ANY,
            config_defaults=expected,
            axapi_args={'server': {}})
        # self.assertIn("member.create", s)
        # self.assertIn(str(expected), s)

    def test_create_expressions_none(self):
        self._test_create_expressions("server", None, {})

    def test_create_expressions_match_beginning(self):
        self._test_create_expressions("secureserver", self.EXPR_BEGIN)

    def test_create_expressions_match_end(self):
        self._test_create_expressions("serverweb", self.EXPR_END)

    def test_create_expressions_match_charclass(self):
        self._test_create_expressions("serverwwserver", self.EXPR_CLASS)

    def test_create_expressions_nomatch(self):
        self.a.config.get_member_expressions = self._get_expressions_mock

        admin_state = self.a.last_client.slb.UP
        m = fake_objs.FakeMember(admin_state_up=admin_state,
                                 pool=mock.MagicMock())

        m.name = "myserver"

        handler = self.a.member
        handler.create(None, m)

        self.a.last_client.slb.server.create.assert_called_with(
            mock.ANY, mock.ANY,
            status=mock.ANY,
            config_defaults={},
            axapi_args={'server': {}})

    def test_create_empty_name_noexception(self):
        self.a.config.get_member_expressions = self._get_expressions_mock

        admin_state = self.a.last_client.slb.UP
        m = fake_objs.FakeMember(admin_state_up=admin_state,
                                 pool=mock.MagicMock())

        m.name = None

        handler = self.a.member
        handler.create(None, m)

        self.a.last_client.slb.server.create.assert_called_with(
            mock.ANY, mock.ANY,
            status=mock.ANY,
            config_defaults={},
            axapi_args={'server': {}})



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import binascii
import logging
import re

import acos_client.errors as acos_errors
import handler_base_v2
import v2_context as a10
# tenant names allow some funky characters; we do not, as of 4.1.0
non_alpha = re.compile('[^0-9a-zA-Z_-]')

LOG = logging.getLogger(__name__)


class MemberHandler(handler_base_v2.HandlerBaseV2):

    def _get_name(self, member, ip_address):
        if self.a10_driver.config.get('member_name_use_uuid'):
            return member.id

        tenant_label = member.tenant_id[:5]
        if non_alpha.search(tenant_label) is not None:
            # This corner-case likely only occurs with silly unit tests
            tenant_label = binascii.hexlify(tenant_label)
        addr_label = str(ip_address).replace(".", "_", 4)
        server_name = "_%s_%s_neutron" % (tenant_label, addr_label)
        return server_name

    def _meta_name(self, member, ip_address):
        return self.meta(member, 'name', self._get_name(member, ip_address))

    def _create(self, c, context, member):
        server_ip = self.neutron.member_get_ip(context, member,
                                               c.device_cfg['use_float'])
        server_name = self._meta_name(member, server_ip)
        conn_limit = c.device_cfg.get('conn-limit')
        status = c.client.slb.UP
        if not member.admin_state_up:
            status = c.client.slb.DOWN
        os_name = member.name

        try:
            server_args = self.meta(member, 'server', {})
            if conn_limit:
                if conn_limit < 1 or conn_limit > 8000000:
                    LOG.warning("The specified member server connection limit " +
                                "(configuration setting: conn-limit) is out of " +
                                "bounds with value {0}. Please set to between " +
                                "1-8000000. Defaulting to 8000000".format(conn_limit))
                else:
                    server_args['conn-limit'] = conn_limit
            server_args = {'server': server_args}
            c.client.slb.server.create(server_name, server_ip,
                                       status=status,
                                       config_defaults=self._get_config_defaults(c, os_name),
                                       axapi_args=server_args)
        except (acos_errors.Exists, acos_errors.AddressSpecifiedIsInUse):
            pass

        try:
            member_args = {'member': self.meta(member, 'member', {})}
            c.client.slb.service_group.member.create(
                self._pool_name(context, pool=member.pool),
                server_name,
                member.protocol_port,
                status=status,
                axapi_args=member_args)
        except acos_errors.Exists:
            pass

    def create(self, context, member):
        with a10.A10WriteStatusContext(self, context, member) as c:
            self._create(c, context, member)
            self.hooks.after_member_create(c, context, member)

    def update(self, context, old_member, member):
        with a10.A10WriteStatusContext(self, context, member) as c:
            server_ip = self.neutron.member_get_ip(context, member,
                                                   c.device_cfg['use_float'])
            server_name = self._meta_name(member, server_ip)

            status = c.client.slb.UP
            if not member.admin_state_up:
                status = c.client.slb.DOWN

            try:
                member_args = {'member': self.meta(member, 'member', {})}
                c.client.slb.service_group.member.update(
                    self._pool_name(context, pool=member.pool),
                    server_name,
                    member.protocol_port,
                    status,
                    axapi_args=member_args)
            except acos_errors.NotFound:
                # Adding db relation after the fact
                self._create(c, context, member)

            self.hooks.after_member_update(c, context, member)

    def _delete(self, c, context, member):
        server_ip = self.neutron.member_get_ip(
            context, member, c.device_cfg['use_float'])
        server_name = self._meta_name(member, server_ip)

        try:
            if self.neutron.member_count(context, member) > 1:
                c.client.slb.service_group.member.delete(
                    self._pool_name(context, pool_id=member.pool_id, pool=member.pool),
                    server_name,
                    member.protocol_port)
            else:
                c.client.slb.server.delete(server_name)
        except acos_errors.NotFound:
            pass

        self.hooks.after_member_delete(c, context, member)

    def delete(self, context, member):
        with a10.A10DeleteContext(self, context, member) as c:
            self._delete(c, context, member)

    def stats(self, context, member):
        retval = {
            "servers_up": 0,
            "servers_down": 0,
            "servers_disable": 0,
            "servers_total": 0
        }

        try:
            with a10.A10Context(self, context, member) as c:
                server_ip = self.neutron.member_get_ip(context, member,
                                                       c.device_cfg['use_float'])
                server_name = self._meta_name(member, server_ip)

                stats = c.client.slb.service_group.member.stats(name=server_name)
                retval = stats

        except Exception as ex:
            LOG.exception(ex)
        finally:
            pass

        return retval

    def _get_expressions(self, c):
        rv = {}
        rv = c.a10_driver.config.get_member_expressions()
        return rv



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
# flake8: noqa

# Version foo

from version import VERSION
from version import VERSION as __version__

# Make sure the '_' function works, as others need it

import gettext
import six

if six.PY2:
    gettext.install('a10', unicode=1)
else:
    gettext.install('a10')

# Automatically export some driver entry points

try:
    import neutron  # noqa
except ImportError:
    # Maybe running tests or utilties?
    pass
else:
    from a10_openstack_lb import A10OpenstackLBV1

try:
    import neutron  # noqa
    import neutron_lbaas  # noqa
except ImportError:
    # Maybe running tests or utilties?
    pass
else:
    from a10_openstack_lb import A10OpenstackLBV2



# Copyright 2016,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.from neutron.db import model_base

import copy
import mock
import os

import alembic.command as alembic_command
import alembic.config as alembic_config
import alembic.op as op
import sqlalchemy

import a10_neutron_lbaas.db.migration as migration
import a10_neutron_lbaas.tests.db.session as session
import test_base


class TestMigrations(test_base.UnitTestBase):

    def setUp(self):
        super(TestMigrations, self).setUp()
        self.patch_create_engine = mock.patch.object(sqlalchemy,
                                                     'create_engine',
                                                     return_value=self.connection)
        self.patch_create_engine.__enter__()

        create_index = op.create_index
        drop_index = op.drop_index

        def create_uniquely_named_index(name, tname, *args, **kwargs):
            return create_index(tname + '_' + name, tname, *args, **kwargs)

        def drop_uniquely_named_index(name, *args, **kwargs):
            return drop_index((kwargs.get('table_name') or args[0]) + '_' + name, *args, **kwargs)

        self.patch_op_create_index = mock.patch.object(op,
                                                       'create_index',
                                                       side_effect=create_uniquely_named_index)
        self.patch_op_drop_index = mock.patch.object(op,
                                                     'drop_index',
                                                     side_effect=drop_uniquely_named_index)
        self.patch_op_create_index.__enter__()
        self.patch_op_drop_index.__enter__()

    def tearDown(self):
        self.patch_op_drop_index.__exit__()
        self.patch_op_create_index.__exit__()
        self.patch_create_engine.__exit__()
        super(TestMigrations, self).tearDown()

    def config(self):
        config = alembic_config.Config(
            os.path.join(os.path.dirname(migration.__file__), 'alembic.ini')
        )
        config.connection = self.connection
        return config

    def upgrade(self, revision):
        alembic_command.upgrade(self.config(), revision)

    def downgrade(self, revision):
        alembic_command.downgrade(self.config(), revision)

    def test_install(self):
        self.upgrade('heads')

    def test_install_schema_matches_model_schema(self):
        self.upgrade('heads')

        a10_models = session.a10_models()

        inspection = sqlalchemy.inspect(self.connection)

        tables = inspection.get_table_names()

        missing_tables = [model.__tablename__ for model in a10_models if
                          model.__tablename__ not in tables]
        self.assertEqual([], missing_tables,
                         "The following tables weren't created by installing {0}".
                         format(missing_tables))

        dialect = self.connection.dialect
        ddl_compiler = dialect.ddl_compiler(dialect, None)

        def normalize(schema_type):
            copied_type = copy.copy(schema_type)
            # We don't care about display width
            if getattr(copied_type, 'display_width', None) is not None:
                copied_type.display_width = None
            if type(schema_type) is sqlalchemy.sql.sqltypes.Text:
                copied_type.length = None

            normalized_type = copied_type.compile(dialect=dialect)

            # mysql has some weird synonyms
            if dialect.name == 'mysql':
                weird_synonyms = {
                    'BOOL': 'TINYINT',
                    'BOOLEAN': 'TINYINT',
                    'TEXT()': 'TEXT'
                }
                normalized_type = weird_synonyms.get(normalized_type, normalized_type)

            return normalized_type

        for model in a10_models:
            columns = inspection.get_columns(model.__tablename__)

            actual_columns = sorted([
                {
                    'nullable': c['nullable'],
                    'default': c['default'],
                    'type': normalize(c['type']),
                    'name': unicode(c['name'])
                }
                for c in columns],
                key=lambda x: x['name'])

            mapper = sqlalchemy.inspect(model)

            expected_columns = sorted([
                {
                    'nullable': c.nullable,
                    'default': ddl_compiler.get_column_default_string(c),
                    'type': normalize(c.type),
                    'name': unicode(c.name)
                }
                for c in mapper.columns
                if c.table.name == model.__tablename__],
                key=lambda x: x['name'])

            if (actual_columns != expected_columns):
                print ("The model and installed columns for {0} don't match".
                       format(model.__tablename__))
                print ("Model columns    ", [c['name'] for c in expected_columns])
                print ("Installed columns", [c['name'] for c in actual_columns])
            self.assertEqual(expected_columns, actual_columns)

    def test_upgrade_heads_downgrade_base(self):
        self.upgrade('heads')
        self.downgrade('base')

    def test_upgrade_heads_downgrade_base_upgrade_heads(self):
        self.upgrade('heads')
        self.downgrade('base')
        self.upgrade('heads')



# Copyright 2016,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

try:
    import neutron.api.v2.attributes as old_attributes
except ImportError:
    old_attributes = None

try:
    import neutron_lib.constants as lib_constants
except ImportError:
    lib_constants = None

try:
    import neutron_lib.api.converters as nlib_converters
except ImportError:
    nlib_converters = None

try:
    import neutron_lib.api.validators as lib_validators
except ImportError:
    lib_validators = None


def _find(*args):
    for f in args:
        try:
            return f()
        except AttributeError:
            pass

convert_to_int = _find(lambda: old_attributes.convert_to_int,
                       lambda: nlib_converters.convert_to_int)

convert_kvp_list_to_dict = _find(lambda: old_attributes.convert_kvp_list_to_dict,
                                 lambda: nlib_converters.convert_kvp_list_to_dict)
convert_to_list = _find(lambda: old_attributes.convert_to_list,
                        lambda: nlib_converters.convert_to_list)
convert_kvp_to_list = _find(lambda: old_attributes.convert_kvp_to_list,
                            lambda: nlib_converters.convert_kvp_to_list)

ATTR_NOT_SPECIFIED = _find(lambda: old_attributes.ATTR_NOT_SPECIFIED,
                           lambda: lib_constants.ATTR_NOT_SPECIFIED)
PLURALS = _find(lambda: old_attributes.PLURALS,
                lambda: {})


def _add_validator_to(validators):
    def add_validator(type_name, validator):
        if type_name not in validators:
            validators[type_name] = validator
    return add_validator


def add_validators(validators):
    """Uniform way to add validators. Call this one!"""
    for type_name, validator in validators.items():
        add_validator(type_name, validator)

validators = _find(lambda: old_attributes.validators,
                   lambda: lib_validators.validators)

# Choose which way to add a validator based on whether validators has moved
add_validator = _find(lambda: _add_validator_to(old_attributes.validators),
                      lambda: lib_validators.add_validator)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import a10_neutron_lbaas.a10_exceptions as a10_ex
import fake_objs
import test_base


def return_one(*args):
    return 1


def return_two(*args):
    return 2


class TestMembers(test_base.UnitTestBase):

    def set_count_1(self):
        self.a.openstack_driver._member_count = return_one

    def set_count_2(self):
        self.a.openstack_driver._member_count = return_two

    def test_get_ip(self):
        m = fake_objs.FakeMember()
        self.a.member.neutron.member_get_ip(None, m, False)
        self.a.openstack_driver._member_get_ip.assert_called_with(
            None, m, False)

    def test_get_name(self):
        m = fake_objs.FakeMember()
        z = self.a.member._get_name(m, '1.1.1.1')
        self.assertEqual(z, '_ten1_1_1_1_1_neutron')

    def test_count(self):
        self.a.member.neutron.member_count(None, fake_objs.FakeMember())

    def _test_create(self, admin_state_up=True, uuid_name=False, conn_limit=800000):
        if uuid_name:
            old = self.a.config.get('member_name_use_uuid')
            self.a.config._config.member_name_use_uuid = True
        m = fake_objs.FakeMember(admin_state_up=admin_state_up)
        ip = self.a.member.neutron.member_get_ip(None, m, True)
        if uuid_name:
            name = m['id']
        else:
            name = self.a.member._get_name(m, ip)
        self.a.member.create(None, m)

        if admin_state_up:
            status = self.a.last_client.slb.UP
        else:
            status = self.a.last_client.slb.DOWN
        if conn_limit < 1 or conn_limit > 8000000:
            self.a.last_client.slb.server.create.assert_called_with(
                name, ip,
                status=status,
                axapi_args={'server': {}})
        else:
            self.a.last_client.slb.server.create.assert_called_with(
                name, ip,
                status=status,
                axapi_args={'server': {'conn-limit': conn_limit}})
        pool_name = self.a.member._pool_name(None, m['pool_id'])
        self.a.last_client.slb.service_group.member.create.assert_called_with(
            pool_name, name, m['protocol_port'], status=status,
            axapi_args={'member': {}})
        if uuid_name:
            self.a.config._config.member_name_use_uuid = old

    def test_create_connlimit(self):
        for k, v in self.a.config.get_devices().items():
            v['conn-limit'] = 1337
        self._test_create(conn_limit=1337)

    def test_create_connlimit_lowerbound(self):
        for k, v in self.a.config.get_devices().items():
            v['conn-limit'] = 8000001
        try:
            self._test_create(conn_limit=8000001)
        except a10_ex.ConnLimitOutOfBounds:
            pass

    def test_create_connlimit_upperbound(self):
        for k, v in self.a.config.get_devices().items():
            v['conn-limit'] = 0
        try:
            self._test_create(conn_limit=0)
        except a10_ex.ConnLimitOutOfBounds:
            pass

    def test_update_down(self):
        m = fake_objs.FakeMember(admin_state_up=False)
        ip = self.a.member.neutron.member_get_ip(None, m, True)
        name = self.a.member._get_name(m, ip)
        self.a.member.update(None, m, m)

        pool_name = self.a.member._pool_name(None, m['pool_id'])
        self.a.last_client.slb.service_group.member.update.assert_called_with(
            pool_name, name, m['protocol_port'],
            self.a.last_client.slb.DOWN,
            axapi_args={'member': {}})

    def test_delete(self):
        m = fake_objs.FakeMember()
        ip = self.a.member.neutron.member_get_ip(None, m, True)

        self.set_count_1()
        self.a.member.delete(None, m)

        self.a.last_client.slb.server.delete(ip)

    def test_delete_count_gt_one(self):
        m = fake_objs.FakeMember()
        ip = self.a.member.neutron.member_get_ip(None, m, True)
        name = self.a.member._get_name(m, ip)

        self.set_count_2()
        self.a.member.delete(None, m)

        pool_name = self.a.member._pool_name(None, m['pool_id'])
        self.a.last_client.slb.service_group.member.delete.assert_called_with(
            pool_name, name, m['protocol_port'])



# Copyright 2015,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import uuid

from a10_openstack_lib.resources import a10_device_instance as a10_device_instance_resources
from neutron.db import common_db_mixin

from a10_neutron_lbaas import a10_config
from a10_neutron_lbaas.db import models
from a10_neutron_lbaas.neutron_ext.common import resources
from a10_neutron_lbaas.neutron_ext.extensions import a10DeviceInstance


LOG = logging.getLogger(__name__)


def _uuid_str():
    return str(uuid.uuid4())


class A10DeviceInstanceDbMixin(common_db_mixin.CommonDbMixin,
                               a10DeviceInstance.A10DeviceInstancePluginBase):

    def __init__(self, *args, **kwargs):
        super(A10DeviceInstanceDbMixin, self).__init__(*args, **kwargs)
        self.config = a10_config.A10Config()

    def _get_a10_device_instance(self, context, a10_device_instance_id):
        try:
            return self._get_by_id(context, models.A10DeviceInstance, a10_device_instance_id)
        except Exception:
            raise a10DeviceInstance.A10DeviceInstanceNotFoundError(a10_device_instance_id)

    def _make_a10_device_instance_dict(self, a10_device_instance_db, fields=None):
        res = {'id': a10_device_instance_db.id,
               'name': a10_device_instance_db.name,
               'description': a10_device_instance_db.description,
               'tenant_id': a10_device_instance_db.tenant_id,
               'username': a10_device_instance_db.username,
               'password': a10_device_instance_db.password,
               'api_version': a10_device_instance_db.api_version,
               'protocol': a10_device_instance_db.protocol,
               'port': a10_device_instance_db.port,
               'autosnat': a10_device_instance_db.autosnat,
               'v_method': a10_device_instance_db.v_method,
               'shared_partition': a10_device_instance_db.shared_partition,
               'use_float': a10_device_instance_db.use_float,
               'default_virtual_server_vrid': a10_device_instance_db.default_virtual_server_vrid,
               'ipinip': a10_device_instance_db.ipinip,
               # Not all device records are nova instances
               'nova_instance_id': a10_device_instance_db.nova_instance_id,
               'host': a10_device_instance_db.host,
               'write_memory': a10_device_instance_db.write_memory}
        return self._fields(res, fields)

    def _get_body(self, a10_device_instance):
        body = a10_device_instance[a10_device_instance_resources.RESOURCE]
        return resources.remove_attributes_not_specified(body)

    def create_a10_device_instance(self, context, a10_device_instance):
        body = self._get_body(a10_device_instance)
        with context.session.begin(subtransactions=True):
            instance_record = models.A10DeviceInstance(
                id=_uuid_str(),
                tenant_id=context.tenant_id,
                name=body.get('name', ''),
                description=body.get('description', ''),
                username=body['username'],
                password=body['password'],
                api_version=body['api_version'],
                protocol=body['protocol'],
                port=body['port'],
                autosnat=body['autosnat'],
                v_method=body['v_method'],
                shared_partition=body['shared_partition'],
                use_float=body['use_float'],
                default_virtual_server_vrid=body['default_virtual_server_vrid'],
                ipinip=body['ipinip'],
                # Not all device records are nova instances
                nova_instance_id=body.get('nova_instance_id'),
                write_memory=body.get('write_memory', False),
                host=body['host'])
            context.session.add(instance_record)

        return self._make_a10_device_instance_dict(instance_record)

    def get_a10_device_instance(self, context, a10_device_instance_id, fields=None):
        instance = self._get_a10_device_instance(context, a10_device_instance_id)
        return self._make_a10_device_instance_dict(instance, fields)

    def get_a10_device_instances(self, context, filters=None, fields=None,
                                 sorts=None, limit=None, marker=None,
                                 page_reverse=False):
        LOG.debug("A10DeviceInstanceDbMixin:get_a10_device_instances() tenant_id=%s" %
                  (context.tenant_id))
        return self._get_collection(context, models.A10DeviceInstance,
                                    self._make_a10_device_instance_dict, filters=filters,
                                    fields=fields, sorts=sorts, limit=limit,
                                    marker_obj=marker, page_reverse=page_reverse)

    def delete_a10_device_instance(self, context, id):
        with context.session.begin(subtransactions=True):
            LOG.debug("A10DeviceInstanceDbMixin:delete_a10_device_instances() id=%s" %
                      (id))
            instance = self._get_by_id(context, models.A10DeviceInstance, id)
            context.session.delete(instance)

    def update_a10_device_instance(self, context, id, a10_device_instance):
        with context.session.begin(subtransactions=True):
            instance = self._get_by_id(context, models.A10DeviceInstance,
                                       id)
            instance.update(**a10_device_instance.get("a10_device_instance"))

            return self._make_a10_device_instance_dict(instance)



# Copyright 2015,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.from neutron.db import model_base

from oslo_log import log as logging

import a10_neutron_lbaas.a10_config as a10_config
from a10_neutron_lbaas.neutron_ext.common import constants
from a10_neutron_lbaas.neutron_ext.common import resources as common_resources

import a10_neutron_lbaas.neutron_ext.db.a10_device_instance as a10_device_instance
import a10_neutron_lbaas.vthunder.instance_manager as instance_manager
from a10_openstack_lib.resources import a10_device_instance as resources

LOG = logging.getLogger(__name__)

# api, vthunder, instance, and db keys
_API = 0
_VTHUNDER_CONFIG = 1
_INSTANCE = 2
_DB = 3
_mappings = [("id", None, None, "id"),
             ("tenant_id", None, None, "tenant_id"),
             ("tenant_id", None, None, "project_id"),
             ("project_id", None, None, "tenant_id"),
             ("project_id", None, None, "project_id"),
             ("name", None, None, "name"),
             ("description", None, None, "description"),
             ("host", None, "ip_address", "host"),
             ("username", "username", None, "username"),
             ("password", "password", None, "password"),
             ("api_version", "api_version", None, "api_version"),
             ("protocol", "protocol", None, "protocol"),
             ("port", "port", None, "port"),
             ("nova_instance_id", None, "nova_instance_id", "nova_instance_id"),
             (None, "autosnat", None, "autosnat"),
             (None, "v_method", None, "v_method"),
             (None, "shared_partition", None, "shared_partition"),
             (None, "use_float", None, "use_float"),
             (None, "default_virtual_server_vrid", None, "default_virtual_server_vrid"),
             (None, "ipinip", None, "ipinip"),
             (None, "write_memory", None, "write_memory"),
             ("management_network", "vthunder_management_network", None, None),
             ("data_networks", "vthunder_data_networks", None, None),
             ("image", "glance_image", None, None),
             ("flavor", "nova_flavor", None, None)]


def _convert(source, from_type, to_type):
    result = {}

    for mapping in _mappings:
        source_key = mapping[from_type]
        if source_key is None or source_key not in source:
            continue

        dest_key = mapping[to_type]
        if dest_key is None:
            continue

        result[dest_key] = source[source_key]

    return result


def _make_api_dict(db_record):
    return _convert(db_record, _DB, _API)


class A10DeviceInstancePlugin(a10_device_instance.A10DeviceInstanceDbMixin):

    supported_extension_aliases = [constants.A10_DEVICE_INSTANCE_EXT]

    def get_a10_device_instances(self, context, filters=None, fields=None):
        LOG.debug(
            "A10DeviceInstancePlugin.get_a10_instances(): filters=%s, fields=%s",
            filters,
            fields)

        db_instances = super(A10DeviceInstancePlugin, self).get_a10_device_instances(
            context, filters=filters, fields=fields)

        return map(_make_api_dict, db_instances)

    def create_a10_device_instance(self, context, a10_device_instance):
        """Attempt to create instance using neutron context"""
        LOG.debug("A10DeviceInstancePlugin.create(): a10_device_instance=%s", a10_device_instance)

        config = a10_config.A10Config()
        vthunder_defaults = config.get_vthunder_config()

        imgr = instance_manager.InstanceManager.from_config(config, context)

        dev_instance = common_resources.remove_attributes_not_specified(
            a10_device_instance.get(resources.RESOURCE))

        # Create the instance with specified defaults.
        vthunder_config = vthunder_defaults.copy()
        vthunder_config.update(_convert(dev_instance, _API, _VTHUNDER_CONFIG))
        instance = imgr.create_device_instance(vthunder_config, dev_instance.get("name"))

        db_record = {}
        db_record.update(_convert(vthunder_config, _VTHUNDER_CONFIG, _DB))
        db_record.update(_convert(dev_instance, _API, _DB))
        db_record.update(_convert(instance, _INSTANCE, _DB))

        # If success, return the created DB record
        # Else, raise an exception because that's what we would do anyway
        db_instance = super(A10DeviceInstancePlugin, self).create_a10_device_instance(
            context, {resources.RESOURCE: db_record})

        return _make_api_dict(db_instance)

    def get_a10_device_instance(self, context, id, fields=None):
        LOG.debug("A10DeviceInstancePlugin.get_a10_instance(): id=%s, fields=%s",
                  id, fields)

        db_instance = super(A10DeviceInstancePlugin, self).get_a10_device_instance(
            context, id, fields=fields)

        return _make_api_dict(db_instance)

    def update_a10_device_instance(self, context, id, a10_device_instance):
        LOG.debug(
            "A10DeviceInstancePlugin.update_a10_device_instance(): id=%s, instance=%s",
            id,
            a10_device_instance)

        db_instance = super(A10DeviceInstancePlugin, self).update_a10_device_instance(
            context,
            id,
            a10_device_instance)

        return _make_api_dict(db_instance)

    def delete_a10_device_instance(self, context, id):
        LOG.debug("A10DeviceInstancePlugin.delete(): id=%s", id)
        # Deleting the actual instance requires knowing the nova instance ID
        instance = super(A10DeviceInstancePlugin, self).get_a10_device_instance(context,
                                                                                id)
        nova_instance_id = instance.get("nova_instance_id")
        config = a10_config.A10Config()
        imgr = instance_manager.InstanceManager.from_config(config, context)
        imgr.delete_instance(nova_instance_id)

        return super(A10DeviceInstancePlugin, self).delete_a10_device_instance(context, id)



# Copyright 2015,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from a10_openstack_lib.resources import a10_device_instance as a10_device_instance_resources

import a10_neutron_lbaas.tests.db.test_base as test_base
import a10_neutron_lbaas.tests.unit.unit_config.helper as unit_config

from a10_neutron_lbaas.neutron_ext.common import constants
from a10_neutron_lbaas.neutron_ext.db import a10_device_instance as a10_device_instance
from a10_neutron_lbaas.neutron_ext.extensions import a10DeviceInstance
from neutron.plugins.common import constants as nconstants


class TestA10DeviceInstanceDbMixin(test_base.UnitTestBase):

    def setUp(self):
        super(TestA10DeviceInstanceDbMixin, self).setUp()
        self._nm_patcher = mock.patch('neutron.manager.NeutronManager')
        nm = self._nm_patcher.start()
        # nm.get_service_plugins.return_value = {
        #    nconstants.LOADBALANCERV2: mock.MagicMock()
        # }

        self._config_cleanup = unit_config.use_config_dir()

        self.plugin = a10_device_instance.A10DeviceInstanceDbMixin()

    def tearDown(self):
        self._config_cleanup()

        super(TestA10DeviceInstanceDbMixin, self).tearDown()

    def context(self):
        session = self.open_session()
        context = mock.Mock(session=session, tenant_id='fake-tenant-id')
        return context

    def envelope(self, body):
        return {a10_device_instance_resources.RESOURCE: body}


class TestA10DeviceInstanceDb(TestA10DeviceInstanceDbMixin):

    def setUp(self):
        super(TestA10DeviceInstanceDb, self).setUp()

    def tearDown(self):
        super(TestA10DeviceInstanceDb, self).tearDown()

    def fake_deviceinstance(self):
        return {
            'name': 'fake-name',
            'description': 'fake-description',
            'host': 'fake-host',
            'api_version': 'fake-version',
            'username': 'fake-username',
            'password': 'fake-password',
            'autosnat': False,
            'default_virtual_server_vrid': None,
            'ipinip': False,
            'use_float': False,
            'v_method': 'LSI',
            'shared_partition': 'shared',
            'write_memory': False,
            'nova_instance_id': 'fake-instance-id',
            'project_id': 'fake-tenant-id',
            'protocol': 'https',
            'port': 442
        }

    def fake_deviceinstance_options(self):
        return {
            'protocol': 'http',
            'port': 12345
        }

    def test_a10_device_instance(self):
        instance = self.fake_deviceinstance()
        context = self.context()
        result = self.plugin.create_a10_device_instance(context, self.envelope(instance))
        context.session.commit()
        self.assertIsNot(result['id'], None)

        expected = {}
        expected.update(instance)
        expected.update(
            {
                'id': result['id'],
                'tenant_id': context.tenant_id,
                'project_id': context.tenant_id

            })
        self.assertEqual(expected, result)

    def test_create_a10_device_instance_options(self):
        instance = self.fake_deviceinstance()
        instance.update(self.fake_deviceinstance_options())
        context = self.context()
        result = self.plugin.create_a10_device_instance(context, self.envelope(instance))
        context.session.commit()
        self.assertIsNot(result['id'], None)
        expected = instance.copy()
        expected.update(
            {
                'id': result['id'],
                'tenant_id': context.tenant_id,
                'project_id': context.tenant_id
            })

        self.assertEqual(expected, result)

    def test_create_a10_device_instance_default_port(self):
        instance = self.fake_deviceinstance()
        instance['port'] = 80
        instance['protocol'] = 'http'
        context = self.context()
        result = self.plugin.create_a10_device_instance(context, self.envelope(instance))
        context.session.commit()
        self.assertIsNot(result['id'], None)
        self.assertEqual(80, result['port'])

    def test_get_a10_device_instance(self):
        instance = self.fake_deviceinstance()
        create_context = self.context()
        create_result = self.plugin.create_a10_device_instance(create_context,
                                                               self.envelope(instance))
        create_context.session.commit()
        context = self.context()
        result = self.plugin.get_a10_device_instance(context, create_result['id'])

        self.assertEqual(create_result, result)

    def test_get_a10_device_instance_not_found(self):
        context = self.context()
        self.assertRaises(
            a10DeviceInstance.A10DeviceInstanceNotFoundError,
            self.plugin.get_a10_device_instance,
            context,
            'fake-deviceinstance-id')

    def test_get_a10_device_instances(self):
        instance = self.fake_deviceinstance()
        create_context = self.context()
        create_result = self.plugin.create_a10_device_instance(create_context,
                                                               self.envelope(instance))
        create_context.session.commit()

        context = self.context()
        result = self.plugin.get_a10_device_instances(context)

        self.assertEqual([create_result], result)

    def test_get_plugin_name(self):
        self.assertIsNot(self.plugin.get_plugin_name(), None)

    def test_get_plugin_description(self):
        self.assertIsNot(self.plugin.get_plugin_description(), None)

    def test_get_plugin_type(self):
        self.assertEqual(self.plugin.get_plugin_type(), constants.A10_DEVICE_INSTANCE)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""a10_tenant_bindings created_at nullable

Revision ID: 3a4bbcea6e1e
Revises: 53ef3417bbaa
Create Date: 2016-08-05 20:17:00.492616

"""

# revision identifiers, used by Alembic.
revision = '3a4bbcea6e1e'
down_revision = '53ef3417bbaa'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.alter_column('a10_tenant_bindings', 'created_at',
                    type_=sa.DateTime, nullable=False,
                    existing_type=sa.DateTime, existing_nullable=True)
    op.alter_column('a10_tenant_bindings', 'updated_at',
                    type_=sa.DateTime, nullable=False,
                    existing_type=sa.DateTime, existing_nullable=True)


def downgrade():
    pass



# Copyright 2015,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import pprint
import time
import uuid

import neutronclient.neutron.client as neutron_client

import novaclient.client as nova_client
import novaclient.exceptions as nova_exceptions

import a10_neutron_lbaas.a10_exceptions as a10_ex
import a10_neutron_lbaas.vthunder.keystone as a10_keystone


pp = pprint.PrettyPrinter(indent=4)


LOG = logging.getLogger(__name__)

CREATE_TIMEOUT = 900

# TODO(mdurrant) - These may need to go into a configuration file.

KEYSTONE_VERSION = "3.0"
NOVA_VERSION = "2.1"
NEUTRON_VERSION = "2.0"
OS_INTERFACE_URLS = ["public", "publicURL"]

_default_server = {
    "id": None,
    "name": None,
    "image": None,
    "flavor": None,
    "meta": {},
    "files": {},
    "min_count": 1,  # optional extension
    "max_count": 1,  # optional extension
    "security_groups": [],
    "userdata": None,
    "key_name": None,  # optional extension
    "availability_zone": None,
    "block_device_mapping": None,  # optional extension
    "block_device_mapping_v2": None,  # optional extension
    "scheduler_hints": {},  # optional extension
    "config_drive": False,  # optional extension
    "disk_config": "AUTO",   # AUTO or MANUAL # optional extension
    "admin_pass": None  # optional extension
}

MISSING_ERR_FORMAT = "{0} with name or id {1} could not be found"


class InstanceManager(object):

    def __init__(self, ks_session, network_ks_session=None,
                 nova_api=None, nova_version=NOVA_VERSION,
                 neutron_api=None):

        # This is the keystone session that we use for spawning instances,
        # aka our "service tenant" user.
        self._ks_session = ks_session

        # And this is the keystone session that we use for finding the network
        # that we are going to plumb into, aka the "end user".
        if network_ks_session is not None:
            self._network_ks_session = network_ks_session
        else:
            self._network_ks_session = ks_session

        # Yes, we really want both of these to use the "service tenant".
        self._nova_api = nova_api or nova_client.Client(
            nova_version, session=self._ks_session)
        self._neutron_api = neutron_api or neutron_client.Client(
            NEUTRON_VERSION, session=self._ks_session)

    @classmethod
    def _factory_with_service_tenant(cls, config, user_keystone_session):
        ks = user_keystone_session

        vth = config.get_vthunder_config()
        if 'service_tenant' in vth:
            service_ks = a10_keystone.KeystoneFromConfig(config)
        else:
            service_ks = ks

        nova_version = config.get('nova_api_version')
        return InstanceManager(
            ks_session=service_ks.session, network_ks_session=ks.session,
            nova_version=nova_version)

    @classmethod
    def from_config(cls, config, openstack_context=None):
        ks = a10_keystone.KeystoneFromContext(config, openstack_context)
        return cls._factory_with_service_tenant(config, ks)

    @classmethod
    def from_cmdline(cls, config, tenant_name, username, password):
        ks = a10_keystone.KeystoneFromPassword(config, tenant_name, username, password)
        return cls._factory_with_service_tenant(config, ks)

    def _build_server(self, instance):
        retval = {}
        for k in _default_server:
            retval[k] = instance.get(k, _default_server[k])

        retval['name'] = retval['name'] or 'a10-' + str(uuid.uuid4())

        return retval

    def list_instances(self, detailed=True, search_opts=None,
                       marker=None, limit=None, sort_keys=None, sort_dirs=None):
        return self._nova_api.servers.list(detailed, search_opts, marker, limit,
                                           sort_keys, sort_dirs)

    def create_instance(self, context):
        return self._create_instance(context)

    def _get_ip_addresses_from_instance(self, addresses, mgmt_network_name):
        address_block = addresses[mgmt_network_name]
        v4addresses = filter(lambda x: x["version"] == 4,
                             address_block)
        return v4addresses[0]["addr"]

    def _create_instance(self, context):
        server = self._build_server(context)

        image_id = context.get("image", None)
        flavor_id = context.get("flavor", None)
        net_ids = context.get("networks")
        image = self.get_image(identifier=image_id)
        flavor = self.get_flavor(identifier=flavor_id)

        networks = self.get_networks(net_ids)

        if image is None:
            raise a10_ex.ImageNotFoundError(MISSING_ERR_FORMAT.format("Image", image_id))

        if flavor is None:
            raise a10_ex.FlavorNotFoundError(MISSING_ERR_FORMAT.format("Flavor", flavor_id))

        if networks is None:
            msg = map(lambda x: MISSING_ERR_FORMAT.format("Network", x), net_ids)
            raise a10_ex.NetworksNotFoundError(msg)

        server["image"] = image.id
        server["flavor"] = flavor.id
        server["nics"] = [{'net-id': x['id']} for x in networks]

        created_instance = self._nova_api.servers.create(**server)

        # Next 6 lines -  Added due to insane API on the other side
        if hasattr(created_instance.manager, 'client'):
            # This craziness works around a bug in Liberty.
            created_instance.manager.client.last_request_id = None
        self._create_server_spinlock(created_instance)

        # Get the IP address of the first interface (should be management)
        ip_address = self._get_ip_addresses_from_instance(
            created_instance.addresses, networks[0]['name'])

        return {
            'name': server['name'],
            'instance': created_instance,
            'ip_address': ip_address,
            'nova_instance_id': created_instance.id
        }

    def _create_server_spinlock(self, created_instance):
        created_id = created_instance.id
        timeout = False
        start_time = time.time()
        sleep_time = 1

        pending_statuses = ["INITALIZED"]
        active_statuses = ["ACTIVE"]
        fatal_statuses = ["ERROR",
                          "SOFT_DELETED",
                          "HARD_DELETED",
                          "STOPPED",
                          "PAUSED"]

        while not timeout:
            get_instance = self._nova_api.servers.get(created_id)
            vm_state = getattr(get_instance, "OS-EXT-STS:vm_state").upper()
            end_time = time.time()
            if ((get_instance.id == created_id and len(get_instance.addresses) > 0
                    and vm_state in active_statuses + pending_statuses)):
                timeout = True
                break
            elif vm_state in fatal_statuses:
                raise Exception("Instance created in error state %s" % (vm_state))
                break

            if end_time - start_time > CREATE_TIMEOUT:
                timeout = True
                raise Exception("Timed out creating instance.")
                break

            time.sleep(sleep_time)

    def delete_instance(self, instance_id):
        try:
            return self._nova_api.servers.delete(instance_id)
        except nova_exceptions.NotFound:
            pass

    def get_instance(self, instance):
        return self._nova_api.servers.get(instance)

    def get_flavor(self, identifier=None):
        result = None
        if identifier is None:
            raise a10_ex.IdentifierUnspecifiedError(
                "Parameter identifier must specify flavor id or name")
        flavor_filter = (lambda x: x is not None and
                         ((hasattr(x, "name") and x.name == identifier)
                          or (hasattr(x, "id") and x.id == identifier)))
        flavors = self._nova_api.flavors.list()
        filtered = filter(flavor_filter, flavors)
        # TODO(mdurrant): What if we accidentally hit multiple flavors?
        if filtered and len(filtered) > 0:
            result = filtered[0]
        return result

    def get_image(self, identifier=None):
        result = None
        images = []
        if identifier is None:
            raise a10_ex.IdentifierUnspecifiedError(
                "Parameter identifier must specify image id or name")
        img_filter = (lambda x: x is not None and
                      ((hasattr(x, "name") and x.name is not None and identifier in x.name)
                       or (hasattr(x, "id") and x.id == identifier)))
        try:
            images = self._nova_api.glance.list()
        except Exception as ex:
            raise a10_ex.ImageNotFoundError(
                "Unable to retrieve images from nova.  Error %s" % (ex))
        filtered = filter(img_filter, images)
        if filtered:
            result = filtered[0]
        return result

    def _handle_missing_networks(self, not_found):
        msg_format = "Network {0} was not found by ID or name."
        msgs = []
        for net in not_found:
            msgs.append(msg_format.format(net))
        ex_msg = "\n".join(msgs)
        LOG.exception(ex_msg)
        raise a10_ex.NetworksNotFoundError(ex_msg)

    def _get_networks(self, session, networks=[]):
        network_list = {"networks": []}
        net_list = []

        if networks is None:
            raise a10_ex.IdentifierUnspecifiedError(
                "Parameter networks must be specified.")

        try:
            # Lookup as user, since names are not unique
            q_api = neutron_client.Client(NEUTRON_VERSION, session=session)
            network_list = q_api.list_networks()
            net_list = network_list.get("networks", [])
        # TODO(mdurrant) - Create specific exceptions.
        except Exception as ex:
            LOG.exception(
                "Unable to retrieve networks from neutron.\nError %s" % (ex))

        # TODO(mdurrant-jk-cshock) - Look up networks by name too
        id_func = (lambda x: x.get("net-id",
                                   x.get("uuid", x.get("id"))) if x is not None else None)

        networks_by_id = dict((id_func(x), x) for x in net_list)
        networks_by_name = dict((x.get("name"), x) for x in net_list)
        available_networks = networks_by_name.copy()
        available_networks.update(networks_by_id)

        missing_networks = [x for x in networks if x not in available_networks.keys()]

        if any(missing_networks):
            self._handle_missing_networks(missing_networks)

        return [{
            'id': id_func(available_networks[x]),
            'name': available_networks[x].get('name', '')
        } for x in networks]

    def get_networks(self, networks=[]):
        if self._ks_session != self._network_ks_session:
            mgmt = self._get_networks(self._ks_session, networks[:1])
            data = self._get_networks(self._network_ks_session, networks[1:])
            return mgmt + data
        else:
            return self._get_networks(self._ks_session, networks)

    def _device_instance(self, vthunder_config, name=None):
        # Pick an image, any image
        image_id = vthunder_config['glance_image']
        if image_id is None:
            raise a10_ex.FeatureNotConfiguredError("Launching instance requires configured image")

        # Get the flavor from config
        flavor = vthunder_config['nova_flavor']
        if flavor is None:
            raise a10_ex.FeatureNotConfiguredError("Launching instance requires configured flavor")

        mgmt_network = vthunder_config.get("vthunder_management_network")

        networks = [mgmt_network] if mgmt_network else []
        networks += vthunder_config.get('vthunder_data_networks')

        if networks is None or len(networks) < 1:
            raise a10_ex.FeatureNotConfiguredError(
                "Launching instance requires configured networks")

        return {
            'name': name,
            'image': image_id,
            'flavor': flavor,
            'networks': networks
        }

    def create_device_instance(self, vthunder_config, name=None):
        instance_configuration = self._device_instance(vthunder_config, name=name)
        return self._create_instance(instance_configuration)

    def _plumb_port(self, server, network_id, wrong_ips):
        """Look for an existing port on the network
        Add one if it doesn't exist
        """

        for attached_interface in server.interface_list():
            if attached_interface.net_id == network_id:
                if any(map(lambda x: x['ip_address'] in wrong_ips, attached_interface.fixed_ips)):
                    continue
                return attached_interface

        return server.interface_attach(None, network_id, None)

    def plumb_instance(self, instance_id, network_id, allowed_ips, wrong_ips=[]):
        server = self._nova_api.servers.get(instance_id)

        interface = self._plumb_port(server, network_id, wrong_ips=wrong_ips)

        port = self._neutron_api.show_port(interface.port_id)

        allowed_address_pairs = port["port"].get("allowed_address_pairs", [])
        existing_allowed_ips = set(pair["ip_address"] for pair in allowed_address_pairs)
        new_allowed_ips = set(allowed_ips) - existing_allowed_ips
        new_address_pairs = [{"ip_address": ip} for ip in new_allowed_ips]

        merged_address_pairs = allowed_address_pairs + new_address_pairs

        self._neutron_api.update_port(interface.port_id, {
            "port": {
                "allowed_address_pairs": merged_address_pairs
            }
        })

        return interface.fixed_ips[0]['ip_address']

    def plumb_instance_subnet(self, instance_id, subnet_id, allowed_ips, wrong_ips=[]):
        subnet = self._neutron_api.show_subnet(subnet_id)
        network_id = subnet["subnet"]["network_id"]
        return self.plumb_instance(instance_id, network_id, allowed_ips, wrong_ips=wrong_ips)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1.identity import v2
from keystoneauth1.identity import v3
from keystoneauth1 import session
from keystoneclient.v2_0 import client as keystone_v2_client
from keystoneclient.v3 import client as keystone_v3_client

import a10_neutron_lbaas.a10_exceptions as a10_ex


class KeystoneBase(object):

    def __init__(self, keystone_version, auth_url, vthunder_config):
        (self._session, self._keystone_client) = (None, None)

    @property
    def session(self):
        return self._session

    @property
    def client(self):
        return self._keystone_client

    def _get_keystone_stuff(self, ks_version, auth):
        sess = session.Session(auth=auth)

        if int(ks_version) == 2:
            ks = keystone_v2_client.Client(session=sess)
        else:
            ks = keystone_v3_client.Client(session=sess)

        return (sess, ks)


class KeystoneFromConfig(KeystoneBase):

    def __init__(self, a10_config):
        vth = a10_config.get_vthunder_config()
        (self._session, self._keystone_client) = self._get_keystone_pw(
            ks_version=a10_config.get('keystone_version'),
            auth_url=a10_config.get('keystone_auth_url'),
            tenant_name=vth['service_tenant']['tenant_name'],
            user=vth['service_tenant']['username'],
            password=vth['service_tenant']['password'])

    def _get_keystone_pw(self, ks_version, auth_url, user, password, tenant_name):
        if int(ks_version) == 2:
            auth = v2.Password(
                auth_url=auth_url, username=user, password=password,
                tenant_name=tenant_name)
        elif int(ks_version) == 3:
            auth = v3.Password(
                auth_url=auth_url, username=user, password=password,
                project_name=tenant_name)
        else:
            raise a10_ex.InvalidConfig('keystone version must be protovol version 2 or 3')

        return self._get_keystone_stuff(ks_version, auth)


class KeystoneFromPassword(KeystoneFromConfig):

    def __init__(self, a10_config, tenant_name, username, password):
        (self._session, self._keystone_client) = self._get_keystone_pw(
            ks_version=a10_config.get('keystone_version'),
            auth_url=a10_config.get('keystone_auth_url'),
            tenant_name=tenant_name, user=username, password=password)


class KeystoneFromContext(KeystoneBase):

    def __init__(self, a10_config, openstack_context):
        (self._session, self._keystone_client) = self._get_keystone_token(
            ks_version=a10_config.get('keystone_version'),
            auth_url=a10_config.get('keystone_auth_url'),
            tenant_id=openstack_context.tenant_id,
            auth_token=openstack_context.auth_token)

    def _get_keystone_token(self, ks_version, auth_url, auth_token, tenant_id):
        if int(ks_version) == 2:
            auth = v2.Token(auth_url=auth_url, token=auth_token, tenant_id=tenant_id)
        elif int(ks_version) == 3:
            auth = v3.Token(auth_url=auth_url, token=auth_token, project_id=tenant_id)
        else:
            raise a10_ex.InvalidConfig('keystone version must be protocol version 2 or 3')

        return self._get_keystone_stuff(ks_version, auth)



# Copyright 2014-2016 A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

import a10_neutron_lbaas.a10_exceptions as a10_ex
import fake_objs
import test_base


class TestVIP(test_base.UnitTestBase):
    def __init__(self, *args):
        super(TestVIP, self).__init__(*args)

    def test_create(self):
        self.a.vip.create(None, fake_objs.FakeVIP())
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('virtual_server.create' in s)
        self.assertTrue('1.1.1.1' in s)
        self.assertTrue('vport.create' in s)
        self.assertTrue('id1' in s)
        self.assertTrue('UP' in s)
        self.a.openstack_driver.plugin.get_pool.assert_called_with(
            None, 'pool1')
        self.assertTrue('HTTP' in s)

    def test_create_pers(self):
        self.a.vip.create(None, fake_objs.FakeVIP('HTTP_COOKIE'))
        s = str(self.a.last_client.mock_calls)
        self.assertTrue("c_pers_name='id1'" in s)

    def test_create_adds_slb(self):
        self.a.vip.create(None, fake_objs.FakeVIP())

    def test_create_unsupported(self):
        try:
            self.a.vip.create(None, fake_objs.FakeVIP('APP_COOKIE'))
        except a10_ex.UnsupportedFeature:
            pass

    def test_create_autosnat_false_v21(self):
        self._test_create_autosnat("2.1", False, "source_nat_auto")

    def test_create_autosnat_true_v21(self):
        self._test_create_autosnat("2.1", True, "source_nat_auto")

    def test_create_autosnat_true_v30(self):
        self._test_create_autosnat("3.0", True, "auto")

    def test_create_autosnat_false_v30(self):
        self._test_create_autosnat("3.0", False, "auto")

    def _test_create_autosnat(self, api_ver=None, autosnat=None, key="auto"):

        """
        Due to how the config is pulled in, we can't override the config
        version here and just expect it to work.
        """
        for k, v in self.a.config.get_devices().items():
            v['api_version'] = api_ver
            v['autosnat'] = autosnat

        vip = fake_objs.FakeVIP()

        self.a.vip.create(None, vip)
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('virtual_server.create' in s)
        self.assertIn('autosnat=%s' % autosnat, s)

    def test_create_default_vrid_none_v21(self):
        self._test_create_default_vrid("2.1", None)

    def test_create_default_vrid_set_v21(self):
        self._test_create_default_vrid("2.1", 7)

    def test_create_default_vrid_none_v30(self):
        self._test_create_default_vrid("3.0", None)

    def test_create_default_vrid_set_v30(self):
        self._test_create_default_vrid("3.0", 7)

    def _test_create_default_vrid(self, api_ver=None, default_vrid=None):

        """
        Due to how the config is pulled in, we override the config
        for all of the devices.
        """

        for k, v in self.a.config.get_devices().items():
            v['api_version'] = api_ver
            v['default_virtual_server_vrid'] = default_vrid

        vip = fake_objs.FakeVIP()
        self.a.vip.create(None, vip)

        create = self.a.last_client.slb.virtual_server.create
        create.assert_has_calls([mock.ANY])
        calls = create.call_args_list

        if default_vrid is not None:
            self.assertIn('vrid=%s' % default_vrid, str(calls))
        if default_vrid is None:
            foundVrid = any(
                'vrid' in x.get('axapi_args', {}).get('virtual_server', {})
                for (_, x) in calls)
            self.assertFalse(
                foundVrid,
                'Expected to find no vrid in {0}'.format(str(calls)))

    def _test_create_ipinip(self, ip_in_ip=False, api_ver="3.0"):
        vip = fake_objs.FakeVIP()

        for k, v in self.a.config.devices.items():
            v['ipinip'] = ip_in_ip
            v['api_version'] = api_ver

        self.a.vip.create(None, vip)
        s = str(self.a.last_client.mock_calls)
        self.assertIn('vport.create', s)
        self.assertIn('ipinip=%s' % ip_in_ip, s)

    def test_create_ip_in_ip_positive_v30(self):
        self._test_create_ipinip(True)

    def test_create_ip_in_ip_negative_v30(self):
        self._test_create_ipinip()

    def test_create_ip_in_ip_positive_v21(self):
        self._test_create_ipinip(True, api_ver="2.1")

    def test_create_ip_in_ip_negative_v21(self):
        self._test_create_ipinip(api_ver="2.1")

    def test_update(self):
        self.a.vip.update(None, fake_objs.FakeVIP(), fake_objs.FakeVIP())
        self.print_mocks()
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('vport.update' in s)
        self.assertTrue('id1' in s)
        self.assertTrue('UP' in s)
        self.a.openstack_driver.plugin.get_pool.assert_called_with(
            None, 'pool1')
        self.assertTrue('HTTP' in s)

    def test_update_delete_pers(self):
        vip_id = "id2"
        self.a.vip.update(None, fake_objs.FakeVIP('SOURCE_IP', vip_id=vip_id), fake_objs.FakeVIP())
        self.print_mocks()
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('vport.update' in s)
        self.assertTrue('id1' in s)
        self.assertTrue('UP' in s)
        self.a.openstack_driver.plugin.get_pool.assert_called_with(None, 'pool1')
        z = self.a.last_client.slb.template.src_ip_persistence.delete
        z.assert_called_with(vip_id)
        self.assertTrue('HTTP' in s)

    def test_update_change_pers(self):
        vip_id = "id2"
        self.a.vip.update(None, fake_objs.FakeVIP('SOURCE_IP', vip_id=vip_id),
                          fake_objs.FakeVIP('HTTP_COOKIE'))
        self.print_mocks()
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('vport.update' in s)
        self.assertTrue('id1' in s)
        self.assertTrue('UP' in s)
        self.a.openstack_driver.plugin.get_pool.assert_called_with(None, 'pool1')
        z = self.a.last_client.slb.template.cookie_persistence.create
        z.assert_called_with("id1")
        self.assertTrue('HTTP' in s)

    def test_delete(self):
        self.a.vip.delete(None, fake_objs.FakeVIP())
        self.a.last_client.slb.virtual_server.delete.assert_called_with('id1')

    def test_delete_removes_slb(self):
        self.a.vip.delete(None, fake_objs.FakeVIP())

    def test_delete_pers(self):
        vip_id = "idx"
        self.a.vip.delete(None, fake_objs.FakeVIP('SOURCE_IP', vip_id=vip_id))
        self.a.last_client.slb.virtual_server.delete.assert_called_with(vip_id)
        z = self.a.last_client.slb.template.src_ip_persistence.delete
        z.assert_called_with(vip_id)

    def _test_create_source_nat_pool(self, source_nat_pool):
        vip = fake_objs.FakeVIP()

        if source_nat_pool:
            for k, v in self.a.config.devices.items():
                v['source_nat_pool'] = source_nat_pool

        self.a.vip.create(None, vip)
        s = str(self.a.last_client.mock_calls)

        self.assertIn("vport.create", s)

        if source_nat_pool:
            expected = "source_nat_pool='{0}'".format(source_nat_pool)
            self.assertIn(expected, s)
        else:
            self.assertIn("source_nat_pool=None", s)

        s = str(self.a.last_client.mock_calls)

    def test_create_source_nat_pool_positive(self):
        self._test_create_source_nat_pool("mypool")

    def test_create_source_nat_pool_negative(self):
        self._test_create_source_nat_pool(None)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import errno
import logging
import time

import acos_client

from a10_neutron_lbaas import a10_exceptions as ex
from a10_neutron_lbaas.db import models
from a10_neutron_lbaas.vthunder import instance_initialization
from a10_neutron_lbaas.vthunder import instance_manager

import base

LOG = logging.getLogger(__name__)


# This next set of plumbing hooks needs to be used when the vthunder
# scheduler is active, for one vthunder per tenant.

class VThunderPerTenantPlumbingHooks(base.BasePlumbingHooks):

    def get_a10_client(self, device_info, **kwargs):
        if kwargs.get('action', None) == 'create':
            retry = [errno.EHOSTUNREACH, errno.ECONNRESET, errno.ECONNREFUSED, errno.ETIMEDOUT]
            return acos_client.Client(
                device_info['host'], device_info['api_version'],
                device_info['username'], device_info['password'],
                port=device_info['port'], protocol=device_info['protocol'],
                retry_errno_list=retry)
        else:
            return super(VThunderPerTenantPlumbingHooks, self).get_a10_client(device_info, **kwargs)

    def _instance_manager(self, a10_context):
        return instance_manager.InstanceManager.from_config(
            self.driver.config, a10_context.openstack_context)

    def _create_instance(self, tenant_id, a10_context, lbaas_obj, db_session):
        start = time.time()
        cfg = self.driver.config
        vth = cfg.get_vthunder_config()
        imgr = self._instance_manager(a10_context)
        instance = imgr.create_device_instance(vth)
        end = time.time()

        LOG.debug("A10 vThunder %s: spawned after %d seconds", instance['nova_instance_id'],
                  end - start)

        from a10_neutron_lbaas.etc import defaults
        device_config = {}
        for key in vth:
            if key in ['status', 'ha_sync_list']:
                continue
            if key in defaults.DEVICE_REQUIRED_FIELDS or key in defaults.DEVICE_OPTIONAL_DEFAULTS:
                device_config[key] = vth[key]
        device_config.update({
            'tenant_id': tenant_id,
            'nova_instance_id': instance['nova_instance_id'],
            'name': instance['name'],
            'host': instance['ip_address'],
        })

        models.A10DeviceInstance.create_and_save(
            db_session=db_session,
            **device_config)

        device_config.update({
            '_perform_initialization': True
        })
        return device_config

    def _wait_for_instance(self, device_config):
        start = time.time()
        client = self.get_a10_client(device_config)
        client.wait_for_connect()
        end = time.time()

        # XXX(dougwig) - this is a <=4.1.0 after CM bug is fixed
        time.sleep(5.0)

        LOG.debug("A10 vThunder %s: ready to connect after %d seconds",
                  device_config['nova_instance_id'], end - start)

    def select_device_with_lbaas_obj(self, tenant_id, a10_context, lbaas_obj,
                                     db_session=None, **kwargs):
        if not self.driver.config.get('use_database'):
            raise ex.RequiresDatabase('vThunder orchestration requires use_database=True')

        # If we already have a vThunder, use it.
        # one vthunder per tenant

        missing_instance = (
            'A10 instance mapped to tenant %s is not present in db; '
            'add it back to config or migrate loadbalancers' % tenant_id
        )

        tb = models.A10TenantBinding.find_by_tenant_id(tenant_id, db_session=db_session)
        if tb is not None:
            d = self.driver.config.get_device(tb.device_name, db_session=db_session)
            if d is None:
                LOG.error(missing_instance)
                raise ex.InstanceMissing(missing_instance)

            LOG.debug("select_device, returning cached instance %s", d)
            return d

        # No? Then we need to create one.

        if kwargs.get('action') != 'create':
            LOG.error(missing_instance)
            raise ex.InstanceMissing(missing_instance)

        device_config = self._create_instance(tenant_id, a10_context, lbaas_obj, db_session)
        self._wait_for_instance(device_config)

        # Now make sure that we remember where it is.

        models.A10TenantBinding.create_and_save(
            tenant_id=tenant_id,
            device_name=device_config['name'],
            db_session=db_session)

        LOG.debug("select_device, returning new instance %s", device_config)
        return device_config

    def after_select_partition(self, a10_context):
        instance = a10_context.device_cfg
        client = a10_context.client

        LOG.debug("after_select_partition, checking instance %s", instance)

        if instance.get('_perform_initialization'):
            instance_initialization.initialize_vthunder(
                a10_context.a10_driver.config, instance, client)

    def after_vip_create(self, a10_context, os_context, vip):
        instance = a10_context.device_cfg
        if 'nova_instance_id' not in instance:
            raise ex.InternalError('Attempting virtual plumbing on non-virtual device')

        if hasattr(vip, 'vip_address'):
            vip_ip_address = vip.vip_address
            vip_subnet_id = vip.vip_subnet_id
        else:
            vip_ip_address = vip['address']
            vip_subnet_id = vip['subnet_id']

        imgr = self._instance_manager(a10_context)
        return imgr.plumb_instance_subnet(
            instance['nova_instance_id'],
            vip_subnet_id,
            [vip_ip_address],
            wrong_ips=[instance['host']])



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import logging
import threading

LATE_INIT_LOCK = threading.Lock()

LOG = logging.getLogger(__name__)


class HandlerBase(object):

    def __init__(self, a10_driver):
        self.a10_driver = a10_driver
        self.openstack_driver = self.a10_driver.openstack_driver

        if self.a10_driver.provider is None:
            with LATE_INIT_LOCK:
                if self.a10_driver.provider is None:
                    for prov, driver in self.openstack_driver.plugin.drivers.items():
                        if driver is self.openstack_driver:
                            self.a10_driver._late_init(prov)
                            break

        self.hooks = a10_driver.hooks

    def _model_type(self):
        return self.__class__.__name__.lower().replace('handler', '')

    def _name(self, obj):
        if isinstance(obj, dict):
            return obj['id']
        else:
            return obj.id

    def _meta_name(self, lbaas_obj):
        return self.meta(lbaas_obj, 'name', self._name(lbaas_obj))

    def _pool_name(self, context, pool_id=None, pool=None):
        if not pool:
            pool = self.neutron.pool_get(context, pool_id)
        if not pool_id:
            pool_id = pool.id
        return self.meta(pool, 'name', pool_id)

    def meta(self, lbaas_obj, key, default):
        if isinstance(lbaas_obj, dict):
            m = lbaas_obj.get('a10_meta', '{}')
        elif hasattr(lbaas_obj, 'a10_meta'):
            m = lbaas_obj.a10_meta
        else:
            return default
        try:
            d = json.loads(m)
        except Exception:
            return default
        return d.get(key, default)



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import a10_neutron_lbaas.handler_base as base
import neutron_ops

import re


class HandlerBaseV2(base.HandlerBase):
    def __init__(self, a10_driver, openstack_manager, neutron=None):
        super(HandlerBaseV2, self).__init__(a10_driver)
        self.openstack_manager = openstack_manager
        if neutron:
            self.neutron = neutron
        else:
            self.neutron = neutron_ops.NeutronOpsV2(self)

    """
    Pass in an element, it's openstack name, and a dictionary of matches.
    """
    def _get_name_matches(self, elem, os_name, redict):
        # for each key in the vport_defaults dictionary
        if not os_name or len(os_name) < 1:
            return

        for k, v in redict.iteritems():
            # check to see if the regex value matches.
            v = redict[k]

            regex_str = v["regex"]
            json_merge = v["json"]
            regex = re.compile(regex_str)
            matched = regex.search(os_name)

            if matched:
                # If so, take those dictionary values and apply them to the object
                elem.update(json_merge)
                break

    def _get_config_defaults(self, c, os_name):
        rv = {}
        # Device-specific defaults have precedence over global
        self._get_name_matches(rv, os_name, self._get_expressions(c))
        return rv



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import acos_client.errors as acos_errors

LOG = logging.getLogger(__name__)


class A10Context(object):

    def __init__(self, handler, openstack_context, openstack_lbaas_obj,
                 **kwargs):
        self.handler = handler
        self.openstack_driver = handler.openstack_driver
        self.a10_driver = handler.a10_driver
        self.hooks = self.a10_driver.hooks
        self.openstack_context = openstack_context
        self.openstack_lbaas_obj = openstack_lbaas_obj
        self.device_name = kwargs.get('device_name', None)
        self.action = kwargs.get('action', '')
        LOG.debug("A10Context obj=%s", openstack_lbaas_obj)
        LOG.debug("A10Context action=%s", self.action)
        self.partition_name = "shared"

    def _get_device(self):
        if self.device_name:
            d = self.a10_driver.config.get_device(self.device_name)
        else:
            d = self.a10_driver._select_a10_device(self.tenant_id, a10_context=self,
                                                   lbaas_obj=self.openstack_lbaas_obj,
                                                   action=self.action)
        return d

    def _get_client(self, device_cfg):
        return self.a10_driver._get_a10_client(device_cfg, action=self.action)

    def __enter__(self):
        self.get_tenant_id()
        self.device_cfg = self._get_device()
        self.client = self._get_client(self.device_cfg)
        self.select_appliance_partition()
        if hasattr(self.hooks, 'after_select_partition'):
            self.hooks.after_select_partition(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            self.client.session.close()
        except acos_errors.InvalidSessionID:
            pass

        if hasattr(self.hooks, 'a10_context_exit_final'):
            self.hooks.a10_context_exit_final(self)

        if exc_type is not None:
            return False

    def get_tenant_id(self):
        if hasattr(self.openstack_lbaas_obj, 'tenant_id'):
            self.tenant_id = self.openstack_lbaas_obj.root_loadbalancer.tenant_id
        else:
            self.tenant_id = self.openstack_lbaas_obj['tenant_id']

    def select_appliance_partition(self):

        name = self.device_cfg.get("shared_partition", "shared")

        if self.device_cfg['v_method'].lower() == 'adp':
            name = self.tenant_id[0:13]

        # If we are not using appliance partitions, we are done.
        if name == 'shared':
            return

        self.partition_name = name

        try:
            self.client.system.partition.active(name)
            return
        except acos_errors.NotFound:
            pass

        # Create it if not found
        self.hooks.partition_create(self.client, self.openstack_context, name)
        self.client.system.partition.active(name)


class A10WriteContext(A10Context):

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None and self.device_cfg.get('write_memory', True):
            try:
                partition_deleted = getattr(self, "partition_deleted", False)
                partition_name = None if partition_deleted else self.partition_name
                self.client.system.action.activate_and_write(partition_name)

            except acos_errors.InvalidSessionID:
                pass

            for v in self.device_cfg.get('ha_sync_list', []):
                self.client.ha.sync(v['ip'], v['username'], v['password'])

        super(A10WriteContext, self).__exit__(exc_type, exc_value, traceback)


class A10ReplayContext(A10WriteContext):

    def __init__(self, *args, **kwargs):
        self._appliance = kwargs.pop('appliance', None)
        super(A10ReplayContext, self).__init__(*args, **kwargs)

    def _get_device(self):
        return self._appliance.device(self)

    def _get_client(self, device_cfg):
        return self._appliance.client(self, device_cfg, action=self.action)


# class A10WriteStatusContext(A10WriteContext):

#     def __exit__(self, exc_type, exc_value, traceback):
#         if exc_type is None:
#             self.openstack_manager.active(self.openstack_context,
#                                           self.openstack_lbaas_obj.id)
#         else:
#             self.openstack_manager.failed(self.openstack_context,
#                                           self.openstack_lbaas_obj.id)

#         super(A10WriteStatusContext, self).__exit__(exc_type, exc_value,
#                                                     traceback)


class A10DeleteContextBase(A10WriteContext):

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            # self.openstack_manager.db_delete(self.openstack_context,
            #                                  self.openstack_lbaas_obj.id)
            self.partition_deleted = False
            self.partition_cleanup_check()

        super(A10DeleteContextBase, self).__exit__(exc_type, exc_value, traceback)

    def remaining_root_objects(self):
        return 1

    def partition_cleanup_check(self):
        # If we are not using appliance partitions, we are done.

        if self.device_cfg['v_method'].lower() != 'adp':
            return

        n = self.remaining_root_objects()
        LOG.debug("A10DeleteContext.partition_cleanup_check(): n=%s" % (n))
        if n == 0:
            try:
                name = self.tenant_id[0:13]
                if not name:
                    return
                self.hooks.partition_delete(self.client, self.openstack_context, name)
                LOG.debug("hooks.partition_delete of %s succeeded " % (name))
                self.partition_deleted = True
            except Exception:
                LOG.exception("A10Driver: partition cleanup failed; ignoring")



# Copyright (C) 2016, A10 Networks Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.from oslo_log.helpers import logging as logging
import logging

from neutron_lbaas.services.loadbalancer import plugin

from a10_neutron_lbaas import constants as certificate_constants
from a10_neutron_lbaas.neutron_ext.common import constants
from a10_neutron_lbaas.neutron_ext.db import certificate_db as certificate_db


LOG = logging.getLogger(__name__)


class A10CertificatePlugin(certificate_db.A10CertificateDbMixin):

    """Implementation of the Neutron SSL Certificate Plugin."""
    supported_extension_aliases = [constants.A10_CERTIFICATE_EXT]

    def __init__(self):
        super(A10CertificatePlugin, self).__init__()
        self.lbplugin = plugin.LoadBalancerPluginv2()

    def get_a10_certificates(self, context, filters=None, fields=None):
        return super(A10CertificatePlugin, self).get_a10_certificates(context, filters, fields)

    def create_a10_certificate(self, context, a10_certificate):
        return super(A10CertificatePlugin, self).create_a10_certificate(context, a10_certificate)

    def get_a10_certificate(self, context, id, fields=None):
        return super(A10CertificatePlugin, self).get_a10_certificate(context, id, fields)

    def update_a10_certificate(self, context, id, a10_certificate):
        return super(A10CertificatePlugin, self).update_a10_certificate(context, id,
                                                                        a10_certificate)

    def delete_a10_certificate(self, context, id):
        return super(A10CertificatePlugin, self).delete_a10_certificate(context, id)

    def get_a10_certificate_bindings(self, context, filters=None, fields=None):
        return super(A10CertificatePlugin, self).get_a10_certificate_bindings(context, filters,
                                                                              fields)

    def _set_a10_certificate_binding_status(self, context, id, status):
        update_binding = {
            "id": id,
            "status": status
        }
        update_a10_certificate_binding = {constants.A10_CERTIFICATE_BINDING: update_binding}

        result = super(A10CertificatePlugin,
                       self).update_a10_certificate_binding(context,
                                                            update_a10_certificate_binding)

        return result

    def _update_listener(self, context, listener_id):
        # Create an empty listener structure - we just want to trigger the update logic
        fake_listener = {"listener": {}}
        # Below will raise exception if listener doesn't exist
        self.lbplugin.update_listener(context, listener_id, fake_listener)

    def create_a10_certificate_binding(self, context, a10_certificate_binding):
        binding = a10_certificate_binding[constants.A10_CERTIFICATE_BINDING]
        created_binding = super(A10CertificatePlugin,
                                self).create_a10_certificate_binding(context,
                                                                     a10_certificate_binding)

        # All of the real work happens in the listener handler.
        self._update_listener(context, binding["listener_id"])

        result = self._set_a10_certificate_binding_status(context, created_binding["id"],
                                                          certificate_constants.STATUS_CREATED)
        return result

    def get_a10_certificate_binding(self, context, id, fields=None):
        return super(A10CertificatePlugin, self).get_a10_certificate_binding(context, id, fields)

    def delete_a10_certificate_binding(self, context, id):
        binding = self._set_a10_certificate_binding_status(context, id,
                                                           certificate_constants.STATUS_DELETING)
        # All of the real work happens in the listener handler.
        # Try to update the listener - it could be gone by now.
        try:
            self._update_listener(context, binding["listener_id"])
        except Exception as ex:
            LOG.exception(ex)
            pass

        return super(A10CertificatePlugin, self).delete_a10_certificate_binding(context, id)



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import acos_client.errors as acos_errors

import handler_base_v2
import v2_context as a10

LOG = logging.getLogger(__name__)


class LoadbalancerHandler(handler_base_v2.HandlerBaseV2):

    def _set(self, set_method, c, context, lb):
        status = c.client.slb.UP
        if not lb.admin_state_up:
            status = c.client.slb.DOWN

        try:
            vip_meta = self.meta(lb, 'virtual_server', {})
            os_name = lb.name
            set_method(
                self._meta_name(lb),
                lb.vip_address,
                status,
                vrid=c.device_cfg.get('default_virtual_server_vrid'),
                config_defaults=self._get_config_defaults(c, os_name),
                axapi_body=vip_meta)
        except acos_errors.Exists:
            pass

    def _create(self, c, context, lb):
        self._set(c.client.slb.virtual_server.create, c, context, lb)

    def _stats_v21(self, c, resp):
        if resp["virtual_server_stat"].get("vport_stat_list"):
            for stat in resp["virtual_server_stat"]["vport_stat_list"]:
                vs = c.client.slb.virtual_service.get(stat["name"])
                if vs["virtual_service"]["service_group"]:
                    pool = c.client.slb.service_group.stats(vs["virtual_service"]["service_group"])
                    stat["pool_stat_list"] = pool["service_group_stat"]

            resp["virtual_server_stat"]["listener_stat"] = resp["virtual_server_stat"].get(
                "vport_stat_list")
            del resp["virtual_server_stat"]["vport_stat_list"]

        resp["loadbalancer_stat"] = resp["virtual_server_stat"]
        del resp["virtual_server_stat"]

        return {
            "bytes_in": resp["loadbalancer_stat"]["req_bytes"],
            "bytes_out": resp["loadbalancer_stat"]["resp_bytes"],
            "active_connections": resp["loadbalancer_stat"]["cur_conns"],
            "total_connections": resp["loadbalancer_stat"]["tot_conns"],
            "extended_stats": resp}

    def _stats_v30(self, c, resp, name):
        stats = {}
        for ports in resp['port-list']:
            for k, v in ports['stats'].items():
                if stats.get(k):
                    stats[k] += v
                else:
                    stats[k] = v

        resp["loadbalancer_stat"] = stats
        resp["loadbalancer_stat"]["listener_stat"] = resp["port-list"]
        del resp["port-list"]

        virt_serv = c.client.slb.virtual_server.get(name)
        for port in virt_serv['virtual-server']['port-list']:
            if port.get("service-group"):
                pool = c.client.slb.service_group.stats(port["service-group"])
                resp["loadbalancer_stat"]["pool_stat_list"] = pool["service-group"]["stats"]
                members = c.client.slb.service_group.get(port["service-group"] + "/member/stats")
                if members:
                    stats = {}
                    for mems in members['member-list']:
                        for k, v in mems['stats'].items():
                            if stats.get(k):
                                stats[k] += v
                            else:
                                stats[k] = v
                    resp["loadbalancer_stat"]["pool_stat_list"].update(stats)
                    resp["loadbalancer_stat"]["pool_stat_list"]["member_list"] = members.get(
                        'member-list')

        return {
            "bytes_in": resp["loadbalancer_stat"]["total_fwd_bytes"],
            "bytes_out": resp["loadbalancer_stat"]["total_rev_bytes"],
            "active_connections": resp["loadbalancer_stat"]["curr_conn"],
            "total_connections": resp["loadbalancer_stat"]["total_conn"],
            "extended_stats": resp
        }

    def create(self, context, lb):
        with a10.A10WriteStatusContext(self, context, lb, action='create') as c:
            self._create(c, context, lb)
            self.hooks.after_vip_create(c, context, lb)

    def update(self, context, old_lb, lb):
        with a10.A10WriteStatusContext(self, context, lb) as c:
            self._set(c.client.slb.virtual_server.update, c, context, lb)
            self.hooks.after_vip_update(c, context, lb)

    def _delete(self, c, context, lb):
        try:
            c.client.slb.virtual_server.delete(self._meta_name(lb))
        except acos_errors.NotFound:
            pass

    def delete(self, context, lb):
        with a10.A10DeleteContext(self, context, lb) as c:
            self._delete(c, context, lb)
            self.hooks.after_vip_delete(c, context, lb)

    def stats(self, context, lb):
        with a10.A10Context(self, context, lb) as c:
            name = self.meta(lb, 'id', lb.id)
            resp = c.client.slb.virtual_server.stats(name)

            if not resp:
                return {
                    "bytes_in": 0,
                    "bytes_out": 0,
                    "active_connections": 0,
                    "total_connections": 0,
                    "extended_stats": {}
                }

            if c.device_cfg.get('api_version') == "3.0":
                return self._stats_v30(c, resp, name)
            else:
                return self._stats_v21(c, resp)

    def refresh(self, context, lb):
        LOG.debug("LB Refresh called.")
        # Ensure all elements associated with this LB exist on the device.

    def _get_expressions(self, c):
        rv = {}
        rv = c.a10_driver.config.get_virtual_server_expressions()
        return rv



# Copyright 2016, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Branch head merge

Revision ID: e3d5134c59a7
Revises: 08dacae1fb67, bc5626a5af2a
Create Date: 2016-12-13 18:33:01.305065

"""

# revision identifiers, used by Alembic.
revision = 'e3d5134c59a7'
down_revision = ('08dacae1fb67', 'bc5626a5af2a')
branch_labels = None
depends_on = None

from alembic import op  # noqa
import sqlalchemy as sa  # noqa


def upgrade():
    pass


def downgrade():
    pass



# Copyright (C) 2016, A10 Networks Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import gettext
_ = gettext.NullTranslations().ugettext

from neutron.db import common_db_mixin as common_db_mixin

import logging
from oslo_utils import uuidutils

from a10_neutron_lbaas import a10_config
from a10_neutron_lbaas.db.models import a10_certificates as models
from a10_neutron_lbaas.neutron_ext.common import constants
from a10_neutron_lbaas.neutron_ext.common import exceptions as nexception
from a10_neutron_lbaas.neutron_ext.extensions import a10Certificate

from a10_openstack_lib.resources import a10_certificate as resources

LOG = logging.getLogger(__name__)


# Certificate exceptions
class CertificateNotFoundError(nexception.NotFound):

    def __init__(self, certificate_id):
        self.msg = _("Certificate {} could not be found.")
        super(CertificateNotFoundError, self).__init__()


class CertificateInUseError(nexception.InUse):

    def __init__(self, certificate_id):
        self.message = _("Certificate is in use and cannot be deleted.")
        self.msg = self.message
        super(CertificateInUseError, self).__init__()


class CertificateVipBindingInUseError(nexception.InUse):
    message = _("Certificate %(certificate_id)s is still bound to %(certificatevipbinding.id)")


class CertificateVipBindingExistsError(nexception.Conflict):
    message = _("Certificate association for Certificate ID: %(certificate_id)s VIP"
                " ID: %(vip_id)s exists.")


class CertificateVipBindingNotFoundByIdError(nexception.NotFound):
    message = _("Certificate associaton for %(id)s could not be found.")


class CertificateVipBindingsNotFoundByCertificateIdError(nexception.NotFound):
    message = _("Certificate associations for %(certificate_id)s could not be found")


class CertificateVipBindingsNotFoundByVipIdError(nexception.NotFound):
    message = _("Certificate associations for %(vip_id)s could not be found")


class CertificateVipBindingsNotFoundByCertificateVipComboError(nexception.NotFound):
    message = _("Certificate associatons for Certificate ID %(certificate_id)s / "
                "VIP ID %(vip_id)s could not be found")


class CertificateListenerBindingInUseError(nexception.InUse):
    message = _(
        "Certificate %(certificate_id)s is still bound to %(certificatelistenerbinding.id)")


class CertificateListenerBindingExistsError(nexception.Conflict):
    message = _("Certificate association for Certificate ID: %(certificate_id)s Listener"
                " ID: %(listener_id)s exists.")


class CertificateListenerBindingNotFoundByIdError(nexception.NotFound):
    message = _("Certificate associaton for %(id)s could not be found.")


class CertificateListenerBindingsNotFoundByCertificateIdError(nexception.NotFound):
    message = _("Certificate associations for %(certificate_id)s could not be found")


class CertificateListenerBindingsNotFoundByListenerIdError(nexception.NotFound):
    message = _("Certificate associations for %(listener_id)s could not be found")


class CertificateListenerBindingsNotFoundByCertificateListenerComboError(nexception.NotFound):
    message = _("Certificate associatons for Certificate ID %(certificate_id)s / "
                "Listener ID %(listener_id)s could not be found")


class A10CertificateDbMixin(common_db_mixin.CommonDbMixin, a10Certificate.A10CertificatePluginBase):

    def __init__(self, *args, **kwargs):
        super(A10CertificateDbMixin, self).__init__(*args, **kwargs)
        self.config = a10_config.A10Config()

    """Class to support SSL certificates and their association with VIPs."""

    # def __init__(self):
    #     # manager = None is used in unit tests where CertificateManager is loaded as a plugin.
    #     pass

    def get_plugin_name(self):
        return constants.A10_CERTIFICATE

    def get_plugin_type(self):
        return constants.A10_CERTIFICATE

    def get_plugin_description(self):
        return "A10 Networks LBaaS Certificates plugin"

    def _get_certificate(self, context, certificate_id):
        try:
            return self._get_by_id(context, models.Certificate, certificate_id)
        except Exception:
            raise CertificateNotFoundError(certificate_id=certificate_id)

    def _make_certificate_dict(self, certificate_db, fields=None):
        res = {'id': certificate_db.id,
               'name': certificate_db.name,
               'tenant_id': certificate_db.tenant_id,
               'description': certificate_db.description,
               'cert_data': certificate_db.cert_data,
               'key_data': certificate_db.key_data,
               'intermediate_data': certificate_db.intermediate_data,
               'password': certificate_db.password}
        return self._fields(res, fields)

    def _ensure_certificate_not_in_use(self, context, certificate_id):
        with context.session.begin(subtransactions=True):
            bindings = (context.session.query(models.CertificateListenerBinding)
                        .filter_by(certificate_id=certificate_id)
                        .all())
            LOG.debug("CertificateDbMixin:_ensure_certificate_not_in_use(): id={0}, len={1}".format(
                certificate_id, len(bindings)))

        if bindings is not None and len(bindings) > 0:
            raise CertificateInUseError(certificate_id)

    def _make_certificate_binding_dict(self, binding, fields=None):
        res = {'id': binding.id,
               'tenant_id': binding.tenant_id,
               'certificate_id': binding.certificate_id,
               'listener_id': binding.listener_id,
               'certificate_name': binding.certificate.name,
               'status': binding.status}
        return self._fields(res, fields)

    def create_a10_certificate(self, context, a10_certificate):
        cert = a10_certificate[resources.CERTIFICATE]

        with context.session.begin(subtransactions=True):
            cert_record = models.Certificate(id=uuidutils.generate_uuid(),
                                             name=cert['name'],
                                             description=cert['description'],
                                             cert_data=cert['cert_data'],
                                             key_data=cert['key_data'],
                                             intermediate_data=cert['intermediate_data'],
                                             password=cert['password'],
                                             tenant_id=context.tenant_id)
            context.session.add(cert_record)

        return self._make_certificate_dict(cert_record)

    def update_a10_certificate(self, context, certificate_id, a10_certificate):
        a10_certificate = a10_certificate[resources.CERTIFICATE]

        with context.session.begin(subtransactions=True):
            certificate_db = self._get_certificate(context, certificate_id)
            certificate_db.update(**a10_certificate)

        return self._make_certificate_dict(certificate_db)

    def delete_a10_certificate(self, context, certificate_id):
        with context.session.begin(subtransactions=True):
            self._ensure_certificate_not_in_use(context, certificate_id)
            LOG.debug("CertificateDbMixin:delete_certificate(): certificate_id={0}".format(
                certificate_id))
            cert = self._get_certificate(context, certificate_id)
            context.session.delete(cert)

    def _get_certificate_binding(self, context, binding_id):
        try:
            return self._get_by_id(context, models.CertificateListenerBinding, binding_id)
        except Exception:
            raise CertificateListenerBindingNotFoundByIdError(id=binding_id)

    def get_a10_certificate(self, context, certificate_id, fields=None):
        cert = self._get_certificate(context, certificate_id)
        return self._make_certificate_dict(cert, fields)

    def get_a10_certificates(self, context, filters=None, fields=None,
                             sorts=None, limit=None, marker=None,
                             page_reverse=False):
        LOG.debug("NDB: CertificateDbMixin:get_certificates() tenant_id=%s" % context.tenant_id)
        return self._get_collection(context, models.Certificate,
                                    self._make_certificate_dict, filters=filters,
                                    fields=fields, sorts=sorts, limit=limit,
                                    marker_obj=marker, page_reverse=page_reverse)

    def get_a10_certificate_binding(self, context, id, fields=None):
        binding = self._get_certificate_binding(context, id)
        LOG.debug("CertificateDbMixin:get_certificate_binding(): %s" % binding)
        return self._make_certificate_binding_dict(binding, fields)

    def create_a10_certificate_binding(self, context, a10_certificate_binding):
        a10_certificate_binding = a10_certificate_binding[resources.CERTIFICATE_BINDING]
        certificate_id = a10_certificate_binding['certificate_id']
        listener_id = a10_certificate_binding['listener_id']
        with context.session.begin(subtransactions=True):
            existing = (context.session.query(models.CertificateListenerBinding)
                        .filter_by(certificate_id=certificate_id, listener_id=listener_id)
                        .first())
            if existing is not None:
                raise CertificateListenerBindingExistsError(certificate_id=certificate_id,
                                                            listener_id=listener_id)
            binding_record = models.CertificateListenerBinding(id=uuidutils.generate_uuid(),
                                                               certificate_id=certificate_id,
                                                               listener_id=listener_id,
                                                               tenant_id=context.tenant_id)
            context.session.add(binding_record)

        return self._make_certificate_binding_dict(binding_record)

    def delete_a10_certificate_binding(self, context, id):
        with context.session.begin(subtransactions=True):
            binding = self._get_certificate_binding(context, id)
            if binding is None:
                raise CertificateListenerBindingNotFoundByIdError(id=id)
            context.session.delete(binding)

    def get_a10_certificate_bindings(self, context, filters=None, fields=None,
                                     sorts=None, limit=None, marker=None,
                                     page_reverse=False):
        bindings = self._get_collection(context, models.CertificateListenerBinding,
                                        self._make_certificate_binding_dict, filters=filters,
                                        fields=fields, sorts=sorts, limit=limit,
                                        marker_obj=marker, page_reverse=page_reverse)
        return bindings

    def get_binding_for_listener(self, context, listener_id):
        with context.session.begin(subtransactions=True):
            existing = (context.session.query(models.CertificateListenerBinding)
                        .filter_by(listener_id=listener_id).first())

        return existing

    def get_bindings_for_certificate(self, context, certificate_id):
        with context.session.begin(subtransactions=True):
            bindings = (context.session.query(models.CertificateListenerBinding)
                        .filter_by(certificate_id=certificate_id).all())

        return bindings

    def update_a10_certificate_binding(self, context, a10_certificate_binding):
        a10_certificate_binding = a10_certificate_binding[resources.CERTIFICATE_BINDING]
        id = a10_certificate_binding["id"]
        with context.session.begin(subtransactions=True):
            certificate_db = self._get_certificate_binding(context, id)
            certificate_db.update(**a10_certificate_binding)

        return self._make_certificate_binding_dict(certificate_db)



vport_expressions = {
   "secure scaleout": {
      'regex': '^secure',
      'json': {
          'scaleout-bucket-count': 128
      }
   }
}

virtual_server_expressions = {
   "described": {
        'regex': '^desc',
        'json': {
            'description': 'Described from the name'
            }
        }
}

service_group_expressions = {
  "mon1": {
    'regex': '^mon',
    'json': {
      "health-check": "mon1"
    }
  }
}

member_expressions = {
  "connections42": {
    'regex': '^connlimit',
    'json': {
        "conn-limit": 4200
    }
  }
}

monitor_expressions = {
    "monitor": {
        "regex": "^tagged",
        "json": {
            "user-tag": 42
        }
    }
}



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import mock

import fake_objs
import test_base

import a10_neutron_lbaas.a10_exceptions as a10_ex
from a10_neutron_lbaas import constants

LOG = logging.getLogger(__name__)


class TestListeners(test_base.HandlerTestBase):

    def test_create_no_lb(self):
        m = fake_objs.FakeListener('TCP', 2222, pool=mock.MagicMock(),
                                   loadbalancer=None)
        try:
            self.a.listener.create(None, m)
        except a10_ex.UnsupportedFeature:
            pass

    def test_create_no_pool(self):
        m = fake_objs.FakeListener('HTTP', 8080, pool=None,
                                   loadbalancer=fake_objs.FakeLoadBalancer())
        self.a.listener.create(None, m)
        self.print_mocks()
        self.assertTrue('create' in str(self.a.last_client.mock_calls))

    def test_create(self):
        admin_states = [True, False]
        persistences = [None, 'SOURCE_IP', 'HTTP_COOKIE', 'APP_COOKIE']
        protocols = ['TCP', 'UDP', 'HTTP', 'HTTPS']
        lb = fake_objs.FakeLoadBalancer()

        for a in admin_states:
            for pers in persistences:
                for p in protocols:
                    self.a.reset_mocks()

                    pool = fake_objs.FakePool(p, 'ROUND_ROBIN', pers)
                    m = fake_objs.FakeListener(p, 2222, pool=pool,
                                               loadbalancer=lb)
                    pool.listener = m
                    saw_exception = False

                    try:
                        self.a.listener.create(None, m)
                    except a10_ex.UnsupportedFeature as e:
                        if pers == 'APP_COOKIE':
                            saw_exception = True
                        else:
                            raise e

                    self.print_mocks()

                    if not saw_exception:
                        s = str(self.a.last_client.mock_calls)
                        self.assertTrue('vport.create' in s)
                        self.assertTrue('fake-lb-id-001' in s)
                        self.assertTrue('fake-listen-id-001' in s)
                        self.assertTrue('port=2222' in s)
                        test_prot = p
                        if p in ('HTTPS', constants.PROTOCOL_TERMINATED_HTTPS):
                            test_prot = 'TCP'
                        self.assertTrue(test_prot in s)

                    if pers == 'SOURCE_IP':
                        self.assertTrue('s_pers_name=None' not in s)
                        pass
                    elif pers == 'HTTP_COOKIE':
                        self.assertTrue('s_pers=HTTP_COOKIE')
                        self.assertTrue('c_pers_name=None' not in s)
                        pass
                    elif pers == 'APP_COOKIE':
                        self.assertTrue('s_pers=APP_COOKIE')
                        self.assertTrue('c_pers_name=None' not in s)
                        pass
                    else:
                        self.assertTrue('c_pers_name=None' in s)
                        self.assertTrue('s_pers_name=None' in s)

    def test_create_autosnat_false_v21(self):
        self._test_create_autosnat("2.1", False)

    def test_create_autosnat_true_v21(self):
        self._test_create_autosnat("2.1", True)

    def test_create_autosnat_unspecified_v21(self):
        self._test_create_autosnat()

    def test_create_autosnat_false_v30(self):
        self._test_create_autosnat("3.0", False)

    def test_create_autosnat_true_v30(self):
        self._test_create_autosnat("3.0", True)

    def test_create_autosnat_unspecified_v30(self):
        self._test_create_autosnat("3.0")

    def _test_create_autosnat(self, api_ver=None, autosnat=None):
        saw_exception = False

        """
        Due to how the config is pulled in, we can't override the config
        version here and just expect it to work.
        """

        for k, v in self.a.config.get_devices().items():
            v['api_version'] = api_ver
            v['autosnat'] = autosnat

        p = 'TCP'
        lb = fake_objs.FakeLoadBalancer()
        pool = fake_objs.FakePool(p, 'ROUND_ROBIN', None)
        m = fake_objs.FakeListener(p, 2222, pool=pool,
                                   loadbalancer=lb)

        try:
            self.a.listener.create(None, m)
        except Exception as e:
            saw_exception = True
            raise e

        if not saw_exception:
            s = str(self.a.last_client.mock_calls)
            self.assertIn('vport.create', s)
            self.assertIn('autosnat=%s' % autosnat, s)

    def _test_create_ipinip(self, api_ver="3.0", ip_in_ip=False):
        for k, v in self.a.config.devices.items():
            v['ipinip'] = ip_in_ip
            v['api_version'] = api_ver

        p = 'TCP'
        lb = fake_objs.FakeLoadBalancer()
        pool = fake_objs.FakePool(p, 'ROUND_ROBIN', None)
        m = fake_objs.FakeListener(p, 2222, pool=pool,
                                   loadbalancer=lb)

        self.a.listener.create(None, m)
        self.print_mocks()

        s = str(self.a.last_client.mock_calls)
        self.assertIn("vport.create", s)
        self.assertIn("ipinip=%s" % ip_in_ip, s)

    def test_create_ip_in_ip_positive_v21(self):
        self._test_create_ipinip(api_ver="2.1", ip_in_ip=True)

    def test_create_ip_in_ip_negative_v21(self):
        self._test_create_ipinip(api_ver="2.1")

    def test_create_ip_in_ip_positive_v30(self):
        self._test_create_ipinip(ip_in_ip=True)

    def test_create_ip_in_ip_negative_v30(self):
        self._test_create_ipinip()

    def test_update_no_lb(self):
        m = fake_objs.FakeListener('TCP', 2222, pool=mock.MagicMock(),
                                   loadbalancer=None)
        try:
            self.a.listener.update(None, m, m)
        except a10_ex.UnsupportedFeature:
            pass

    def test_update_no_pool(self):
        m = fake_objs.FakeListener('HTTP', 8080, pool=None,
                                   loadbalancer=fake_objs.FakeLoadBalancer())
        self.a.listener.create(None, m)
        self.assertFalse('update' in str(self.a.last_client.mock_calls))

    def test_update(self):
        pool = fake_objs.FakePool('HTTP', 'ROUND_ROBIN', None)
        lb = fake_objs.FakeLoadBalancer()
        m = fake_objs.FakeListener('HTTP', 2222, pool=pool, loadbalancer=lb)
        pool.listener = m

        self.a.listener.update(None, m, m)

        self.print_mocks()
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('vport.update' in s)
        self.assertTrue('fake-lb-id-001' in s)
        self.assertTrue('fake-listen-id-001' in s)
        self.assertTrue('port=2222' in s)
        self.assertTrue('HTTP' in s)

    def test_delete(self):
        pool = fake_objs.FakePool('HTTP', 'ROUND_ROBIN', None)
        lb = fake_objs.FakeLoadBalancer()
        m = fake_objs.FakeListener('HTTP', 2222, pool=pool, loadbalancer=lb)
        pool.listener = m

        self.a.listener.delete(None, m)

        self.print_mocks()
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('vport.delete' in s)
        self.assertTrue('fake-lb-id-001' in s)
        self.assertTrue('fake-listen-id-001' in s)
        self.assertTrue('port=2222' in s)
        self.assertTrue('HTTP' in s)

    def test_delete_removes_bindings(self):
        pool = fake_objs.FakePool('TCP', 'ROUND_ROBIN', None)
        lb = fake_objs.FakeLoadBalancer()
        m = fake_objs.FakeListener('TCP', 2222, pool=pool, loadbalancer=lb)
        binding = fake_objs.FakeCertificateBinding(listener_id=m.id)
        pool.listener = m
        bindings = [binding]
        get_bindings_mock = mock.Mock(return_value=bindings)
        self.a.listener.cert_db.get_bindings_for_listener = get_bindings_mock
        self.a.config.use_database = True
        self.a.listener.delete(None, m)
        expected = self.a.listener.cert_db.delete_a10_certificate_binding.call_count
        self.assertEqual(expected, len(bindings))

    def test_delete_no_db_delete_use_database_false(self):
        pool = fake_objs.FakePool('TCP', 'ROUND_ROBIN', None)
        lb = fake_objs.FakeLoadBalancer()
        m = fake_objs.FakeListener('TCP', 2222, pool=pool, loadbalancer=lb)
        binding = fake_objs.FakeCertificateBinding(listener_id=m.id)
        pool.listener = m
        bindings = [binding]
        get_bindings_mock = mock.Mock(return_value=bindings)
        self.a.listener.cert_db.get_bindings_for_listener = get_bindings_mock
        self.a.listener.delete(None, m)
        expected = self.a.listener.cert_db.delete_a10_certificate_binding.call_count
        self.assertEqual(expected, 0)

    def _test_create_source_nat_pool(self, source_nat_pool):
        if source_nat_pool:
            for k, v in self.a.config.devices.items():
                v['source_nat_pool'] = source_nat_pool
        p = 'TCP'
        lb = fake_objs.FakeLoadBalancer()
        pool = fake_objs.FakePool(p, 'ROUND_ROBIN', None)
        m = fake_objs.FakeListener(p, 2222, pool=pool,
                                   loadbalancer=lb)

        self.a.listener.create(None, m)
        self.print_mocks()

        s = str(self.a.last_client.mock_calls)

        self.assertIn("vport.create", s)

        if source_nat_pool:
            expected = "source_nat_pool='{0}'".format(source_nat_pool)
            self.assertIn(expected, s)
        else:
            self.assertIn("source_nat_pool=None", s)

    def test_create_source_nat_pool_positive(self):
        self._test_create_source_nat_pool("mypool")

    def test_create_source_nat_pool_negative(self):
        self._test_create_source_nat_pool(None)

    def _set_device_config(self, key, value):
        for k, v in self.a.config.devices.items():
            v[key] = value

    def _get_vport_config(self):
        return {
            "gslb-enable": 1
        }

    def test_create_vport_defaults_device(self):
        expected = self._get_vport_config()
        self._set_device_config("vport_defaults", expected)

        p = 'TCP'
        lb = fake_objs.FakeLoadBalancer()
        pool = fake_objs.FakePool(p, 'ROUND_ROBIN', None)
        m = fake_objs.FakeListener(p, 2222, pool=pool,
                                   loadbalancer=lb)

        self.a.listener.create(None, m)
        self.print_mocks()

        s = str(self.a.last_client.mock_calls)

        self.assertIn("vport.create", s)
        self.assertIn(str(expected), s)

    def test_create_vport_defaults_global(self):
        expected = self.a.config.get_vport_defaults()

        p = 'TCP'
        lb = fake_objs.FakeLoadBalancer()
        pool = fake_objs.FakePool(p, 'ROUND_ROBIN', None)
        m = fake_objs.FakeListener(p, 2222, pool=pool,
                                   loadbalancer=lb)
        handler = self.a.listener
        handler.create(None, m)
        self.print_mocks()

        s = str(self.a.last_client.mock_calls)

        self.assertIn("vport.create", s)
        self.assertIn(str(expected), s)

    def _test_create_expressions(self, os_name, pattern, expressions=None):
        self.a.config.get_vport_expressions = self._get_expressions_mock
        expressions = expressions or self.a.config.get_vport_expressions()
        expected = expressions.get(pattern, {}).get("json", {})
        p = 'TCP'
        lb = fake_objs.FakeLoadBalancer()
        pool = fake_objs.FakePool(p, 'ROUND_ROBIN', None)
        m = fake_objs.FakeListener(p, 2222, pool=pool,
                                   loadbalancer=lb)
        m.name = os_name
        handler = self.a.listener
        handler.create(None, m)

        s = str(self.a.last_client.mock_calls)
        self.assertIn("vport.create", s)
        self.assertIn(str(expected), s)

    def test_create_vport_expressions_none(self):
        p = 'TCP'
        lb = fake_objs.FakeLoadBalancer()
        pool = fake_objs.FakePool(p, 'ROUND_ROBIN', None)
        m = fake_objs.FakeListener(p, 2222, pool=pool,
                                   loadbalancer=lb)
        handler = self.a.listener
        handler.create(None, m)

        s = str(self.a.last_client.mock_calls)
        self.assertIn("vport.create", s)

    def test_create_vport_expressions_match_beginning(self):
        self._test_create_expressions("securelistener", self.EXPR_BEGIN)

    def test_create_vport_expressions_match_end(self):
        self._test_create_expressions("blahweb", self.EXPR_END)

    def test_create_vport_expressions_match_charclass(self):
        self._test_create_expressions("listenerwwwlis", self.EXPR_CLASS)

    def test_create_vport_expressions_nomatch(self):
        self.a.config.get_vport_expressions = self._get_expressions_mock
        expressions = self.a.config.get_vport_expressions()

        expected = expressions[self.EXPR_BEGIN].get("json", {})
        p = 'TCP'
        lb = fake_objs.FakeLoadBalancer()
        pool = fake_objs.FakePool(p, 'ROUND_ROBIN', None)
        m = fake_objs.FakeListener(p, 2222, pool=pool,
                                   loadbalancer=lb)
        m.name = "insecurelistener"
        handler = self.a.listener
        handler.create(None, m)

        s = str(self.a.last_client.mock_calls)
        self.assertIn("vport.create", s)
        self.assertNotIn(str(expected), s)

    def test_create_vport_expressions_empty(self):
        get_expr_mock = mock.MagicMock()
        get_expr_mock.return_value = {}
        self.a.config.get_vport_expressions = get_expr_mock

        p = 'TCP'
        lb = fake_objs.FakeLoadBalancer()
        pool = fake_objs.FakePool(p, 'ROUND_ROBIN', None)
        m = fake_objs.FakeListener(p, 2222, pool=pool, loadbalancer=lb)
        m.name = "insecurelistener"
        handler = self.a.listener
        handler.create(None, m)
        # This test should just run without raising any exceptions

    def test_create_vport_expressions_noname(self):
        get_expr_mock = mock.MagicMock()
        get_expr_mock.return_value = {}
        self.a.config.get_vport_expressions = get_expr_mock

        p = 'TCP'
        lb = fake_objs.FakeLoadBalancer()
        pool = fake_objs.FakePool(p, 'ROUND_ROBIN', None)
        m = fake_objs.FakeListener(p, 2222, pool=pool, loadbalancer=lb)
        m.name = None
        handler = self.a.listener
        handler.create(None, m)
        # This test should just run without raising any exceptions



# Copyright (C) 2015, A10 Networks Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest


def assertIn(expected, actual):
    if expected not in actual:
        raise Exception("Expected to find {0} in {1}".format(expected, actual))


def assertIsNot(a, b):
    if a is b:
        raise Exception("Expected {0} to not be {1}".format(a, b))


class TestCase(unittest.TestCase):
    """unittest.TestCase with portable or custom assertions"""

    def __init__(self, *args):
        super(TestCase, self).__init__(*args)

        self._patch("assertIn", assertIn)
        self._patch("assertIsNot", assertIsNot)

    def _patch(self, key, value):
        if not hasattr(self, key):
            setattr(self, key, value)



# Copyright 2015 A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import mock

import fake_objs
import test_base

import a10_neutron_lbaas.a10_exceptions as a10_ex
from a10_neutron_lbaas import constants


LOG = logging.getLogger(__name__)


class TestListenersTerminatedHTTPS(test_base.UnitTestBase):

    def test_update_no_lb(self):
        m = fake_objs.FakeListener('TCP', 2222, pool=mock.MagicMock(),
                                   loadbalancer=None)
        try:
            self.a.listener.update(None, m, m)
        except a10_ex.UnsupportedFeature:
            pass

    def test_update_no_pool(self):
        m = fake_objs.FakeListener('HTTP', 8080, pool=None,
                                   loadbalancer=fake_objs.FakeLoadBalancer())
        self.a.listener.create(None, m)
        self.assertFalse('update' in str(self.a.last_client.mock_calls))

    def test_update(self):
        pool = fake_objs.FakePool(constants.PROTOCOL_TERMINATED_HTTPS,
                                  constants.LB_METHOD_ROUND_ROBIN, None)
        lb = fake_objs.FakeLoadBalancer()
        m = fake_objs.FakeListener(constants.PROTOCOL_TERMINATED_HTTPS, 2222,
                                   pool=pool, loadbalancer=lb)
        certmgr = FakeCertManager()

        self.a.barbican_client = certmgr
        # self.a.listener.set_certmgr(certmgr)

        pool.listener = m

        self.a.listener.update(None, m, m)

        self.print_mocks()
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('vport.update' in s)
        self.assertTrue('fake-lb-id-001' in s)
        self.assertTrue('fake-listen-id-001' in s)
        self.assertTrue('port=2222' in s)
        self.assertTrue('HTTPS' in s)

    def test_delete(self):
        pool = fake_objs.FakePool(constants.PROTOCOL_TERMINATED_HTTPS,
                                  constants.LB_METHOD_ROUND_ROBIN, None)
        lb = fake_objs.FakeLoadBalancer()
        m = fake_objs.FakeListener(constants.PROTOCOL_TERMINATED_HTTPS, 2222,
                                   pool=pool, loadbalancer=lb)

        pool.listener = m

        self.a.listener.delete(None, m)

        self.print_mocks()
        s = str(self.a.last_client.mock_calls)
        LOG.debug("DELETE RESULT %s" % s)
        self.assertTrue('vport.delete' in s)
        self.assertTrue('fake-lb-id-001' in s)
        self.assertTrue('fake-listen-id-001' in s)
        self.assertTrue('port=2222' in s)
        self.assertTrue('HTTPS' in s)

    def test_create_protocol_terminated_https(self):
        pool = fake_objs.FakePool(constants.PROTOCOL_TERMINATED_HTTPS,
                                  constants.LB_METHOD_ROUND_ROBIN, None)
        lb = fake_objs.FakeLoadBalancer()
        m = fake_objs.FakeListener(constants.PROTOCOL_TERMINATED_HTTPS, 2222,
                                   pool=pool, loadbalancer=lb)
        m.default_tls_container_id = "CONTAINERID"
        pool.listener = m

        certmgr = FakeCertManager()

        self.a.barbican_client = certmgr
        self.a.listener.create(None, m)
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('HTTPS' in s)

    @mock.patch('a10_neutron_lbaas.v2.handler_listener.certwrapper')
    def test_barbican_client_not_null(self, certmgr):
        """This tests that the barbican_client dependency is always there."""
        handler = self.a.listener

        pool = fake_objs.FakePool(constants.PROTOCOL_TERMINATED_HTTPS,
                                  constants.LB_METHOD_ROUND_ROBIN, None)
        lb = fake_objs.FakeLoadBalancer()
        m = fake_objs.FakeListener(constants.PROTOCOL_TERMINATED_HTTPS, 2222,
                                   pool=pool, loadbalancer=lb)
        pool.listener = m
        handler.create(None, m)
        self.assertTrue(handler.barbican_client is not None)

    def _tls_container_shared(self):
        certmgr = FakeCertManager()
        certmgr.certificate = "CERTDATA"
        certmgr.private_key = "PRIVATEKEY"
        certmgr.private_key_passphrase = "SECRETPASSWORD"

        self.a.barbican_client = certmgr
        handler = self.a.listener

        pool = fake_objs.FakePool(constants.PROTOCOL_TERMINATED_HTTPS,
                                  constants.LB_METHOD_ROUND_ROBIN, None)
        lb = fake_objs.FakeLoadBalancer()
        m = fake_objs.FakeListener(constants.PROTOCOL_TERMINATED_HTTPS, 2222,
                                   pool=pool, loadbalancer=lb)
        return (certmgr, m, pool, handler)

    def test_create_tls_container_positive(self):
        """Test that cert data is passed to handler when data is available"""
        certmgr, listener, pool, handler = self._tls_container_shared()
        listener.default_tls_container_id = "CONTAINERID"

        pool.listener = listener

        handler.create(None, listener)
        s = str(self.a.last_client.mock_calls)

        self.assertIn(certmgr.certificate, s)
        self.assertIn(certmgr.private_key, s)
        self.assertIn(certmgr.private_key_passphrase, s)

    def test_create_tls_container_negative(self):
        """Test that cert data is not passed to handler"""

        certmgr, listener, pool, handler = self._tls_container_shared()
        listener.default_tls_container_id = None

        pool.listener = listener
        handler.create(None, listener)
        s = str(self.a.last_client.mock_calls)
        self.assertFalse("cert=" in s)
        self.assertFalse("key=" in s)


class FakeCertManager(object):

    def __init__(self):
        self.private_key = ""
        self.certificate = ""
        self.private_key_passphrase = ""
        self.container_name = "tls-container"

        self.mock_cert_container = mock.Mock()
        self.mock_cert_container.configure_mock(name=self.container_name)

        self.mock_certificate = mock.Mock(
            get_certificate=lambda: self.certificate,
            get_private_key=lambda: self.private_key,
            get_private_key_passphrase=lambda: self.private_key_passphrase,
            _cert_container=self.mock_cert_container)

    def get_certificate(self, container_id, project_id=None, **kwargs):

        # project_id is needed to call a CertManager in Mitaka and later
        if project_id is None:
            raise Exception("project_id is required")

        return self.mock_certificate



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sqlalchemy as sa

from a10_neutron_lbaas.db import model_base


class A10DeviceInstance(model_base.A10BaseMixin, model_base.A10Base):
    """An orchestrated vThunder that is being used as a device."""

    __tablename__ = 'a10_device_instances'

    # This field is directly analagous to the device name in config.py;
    # and will be used as such throughout.
    name = sa.Column(sa.String(1024), nullable=False)
    description = sa.Column(sa.String(255), nullable=True)

    username = sa.Column(sa.String(255), nullable=False)
    password = sa.Column(sa.String(255), nullable=False)

    api_version = sa.Column(sa.String(12), nullable=False)
    protocol = sa.Column(sa.String(32), nullable=False)
    port = sa.Column(sa.Integer, nullable=False)
    autosnat = sa.Column(sa.Boolean(), nullable=False)
    v_method = sa.Column(sa.String(32), nullable=False)
    shared_partition = sa.Column(sa.String(1024), nullable=False)
    use_float = sa.Column(sa.Boolean(), nullable=False)
    default_virtual_server_vrid = sa.Column(sa.Integer, nullable=True)
    ipinip = sa.Column(sa.Boolean(), nullable=False)
    write_memory = sa.Column(sa.Boolean(), nullable=False)

    nova_instance_id = sa.Column(sa.String(36), nullable=False)
    host = sa.Column(sa.String(255), nullable=False)

    # TODO(dougwig) -- later - reference to scheduler, or capacity, or?
    # TODO(dougwig) -- later - should add state enum here

    # For "device" dicts, use a10_config.get_device()
    # For client objects, use _get_a10_client with the a10_config device dict



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from a10_neutron_lbaas.acos import openstack_mappings
import acos_client.errors as acos_errors
import handler_base_v1
import v1_context as a10

LOG = logging.getLogger(__name__)


class PoolHandler(handler_base_v1.HandlerBaseV1):

    def _set(self, set_method, c, context, pool):
        args = {'service_group': self.meta(pool, 'service_group', {})}

        set_method(
            self._meta_name(pool),
            protocol=openstack_mappings.service_group_protocol(c, pool['protocol']),
            lb_method=openstack_mappings.service_group_lb_method(c, pool['lb_method']),
            axapi_args=args)

    def create(self, context, pool):
        with a10.A10WriteStatusContext(self, context, pool, action='create') as c:
            try:
                self._set(c.client.slb.service_group.create,
                          c, context, pool)
            except acos_errors.Exists:
                pass

    def update(self, context, old_pool, pool):
        # id_func = lambda x: x.get("monitor_id")

        with a10.A10WriteStatusContext(self, context, pool) as c:
            self._set(c.client.slb.service_group.update,
                      c, context, pool)

    def delete(self, context, pool):
        with a10.A10DeleteContext(self, context, pool) as c:
            pool_id = pool.get("id")

            for member in pool['members']:
                m = self.neutron.member_get(context, member)
                self.a10_driver.member._delete(c, context, m)

            for hm in pool.get('health_monitors_status', []):
                z = self.neutron.hm_get(context, hm['monitor_id'])
                self.a10_driver.hm.dissociate(c, context, z, pool_id)

            if 'vip_id' in pool and pool['vip_id'] is not None:
                vip = self.neutron.vip_get(context, pool['vip_id'])
                self.a10_driver.vip._delete(c, context, vip)

            try:
                c.client.slb.service_group.delete(self._meta_name(pool))
            except (acos_errors.NotFound, acos_errors.NoSuchServiceGroup):
                pass

    def stats(self, context, pool_id):
        tenant_id = self.neutron.pool_get_tenant_id(context, pool_id)
        pool = {'id': pool_id, 'tenant_id': tenant_id}
        with a10.A10Context(self, context, pool) as c:
            try:
                vip_id = self.neutron.vip_get_id(context, pool['id'])
                vip = self.neutron.vip_get(context, vip_id)
                name = self.meta(vip, 'vip_name', vip['id'])
                r = c.client.slb.virtual_server.stats(name)
                return {
                    "bytes_in": r["virtual_server_stat"]["req_bytes"],
                    "bytes_out": r["virtual_server_stat"]["resp_bytes"],
                    "active_connections":
                        r["virtual_server_stat"]["cur_conns"],
                    "total_connections": r["virtual_server_stat"]["tot_conns"]
                }
            except Exception:
                return {
                    "bytes_in": 0,
                    "bytes_out": 0,
                    "active_connections": 0,
                    "total_connections": 0
                }



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sqlalchemy as sa

from a10_neutron_lbaas.db import model_base


class A10TenantBinding(model_base.A10BaseMixin, model_base.A10Base):
    __tablename__ = "a10_tenant_bindings"

    device_name = sa.Column(sa.String(1024), nullable=False)

    # XXX(dougwig) -- remove this in favor of find_by()
    @classmethod
    def find_by_tenant_id(cls, tenant_id, db_session=None):
        return cls.find_by_attribute('tenant_id', tenant_id, db_session)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""a10_device_instance description

Revision ID: c4e1caaa618d
Revises: 78d283c111ac
Create Date: 2017-02-23 15:06:26.630570

"""

# revision identifiers, used by Alembic.
revision = 'c4e1caaa618d'
down_revision = '78d283c111ac'
branch_labels = None
depends_on = None

from alembic import op  # noqa
import sqlalchemy as sa  # noqa


def upgrade():
    op.add_column('a10_device_instances',
                  sa.Column('description', sa.String(255), nullable=True)
                  )


def downgrade():
    op.drop_column('a10_device_instances', 'description')



# Copyright 2015,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.from neutron.db import model_base

from neutron.plugins.common import constants as nconstants

from a10_openstack_lib.resources import a10_certificate
from a10_openstack_lib.resources import a10_device_instance

A10_DEVICE_INSTANCE_EXT = a10_device_instance.EXTENSION
A10_DEVICE_INSTANCE = a10_device_instance.SERVICE

A10_CERTIFICATE_EXT = a10_certificate.EXTENSION
A10_CERTIFICATE = a10_certificate.SERVICE
A10_CERTIFICATE_BINDING = a10_certificate.CERTIFICATE_BINDING

nconstants.EXT_TO_SERVICE_MAPPING[A10_CERTIFICATE_EXT] = A10_CERTIFICATE
nconstants.EXT_TO_SERVICE_MAPPING[A10_DEVICE_INSTANCE_EXT] = A10_DEVICE_INSTANCE
try:
    nconstants.ALLOWED_SERVICES.append(A10_CERTIFICATE)
    nconstants.ALLOWED_SERVICES.append(A10_DEVICE_INSTANCE)

    nconstants.COMMON_PREFIXES[A10_CERTIFICATE] = ""
    nconstants.COMMON_PREFIXES[A10_DEVICE_INSTANCE] = ""
except AttributeError:
    # In Liberty and later, ALLOWED_SERVICES is derived from EXT_TO_SERVICE_MAPPING
    # COMMON_PREFIXES are instead gotten from plugin.path_prefix
    pass



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import a10_neutron_lbaas.tests.unit.test_base as test_base


class UnitTestBase(test_base.UnitTestBase):

    def __init__(self, *args):
        super(UnitTestBase, self).__init__(*args)
        self.version = 'v1'



# Copyright 2014, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class MonkeyPatch(object):

    def __init__(self, plugin):
        self.plugin = plugin

    def stats(self, context, loadbalancer_id):
        lb = self.plugin.db.get_loadbalancer(context, loadbalancer_id)
        driver = self.plugin._get_driver_for_loadbalancer(context, loadbalancer_id)
        stats_data = driver.load_balancer.stats(context, lb)
        if stats_data:
            self.plugin.db.update_loadbalancer_stats(context, loadbalancer_id,
                                                     stats_data)
        db_stats = self.plugin.db.stats(context, loadbalancer_id)
        setattr(db_stats, "extended_stats", stats_data['extended_stats'])
        return {'stats': db_stats.to_api_dict()}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from contextlib import contextmanager
import datetime
import uuid

import sqlalchemy as sa
from sqlalchemy.inspection import inspect

from a10_neutron_lbaas.db import api as db_api

Base = db_api.get_base()


def _uuid_str():
    return str(uuid.uuid4())


def _get_date():
    return datetime.datetime.now()


class A10Base(Base):
    __abstract__ = True

    @classmethod
    @contextmanager
    def _query(cls, db_session=None):
        with db_api.magic_session(db_session) as db:
            yield db.query(cls)

    @classmethod
    def get(cls, key, db_session=None):
        with cls._query(db_session) as q:
            return q.get(key)

    @classmethod
    def find_all_by(cls, db_session=None, **kwargs):
        with cls._query(db_session) as q:
            return q.filter_by(**kwargs).all()

    @classmethod
    def find_by(cls, db_session=None, **kwargs):
        with cls._query(db_session) as q:
            return q.filter_by(**kwargs).first()

    @classmethod
    def find_by_attribute(cls, attribute_name, attribute, db_session=None):
        with cls._query(db_session) as q:
            return q.filter(
                getattr(cls, attribute_name) == attribute).first()

    @classmethod
    def find_all(cls, db_session=None):
        with cls._query(db_session) as q:
            return q.all()

    @classmethod
    def create(cls, **kwargs):
        instance = cls(**kwargs)
        # Populate all the unspecified columns with their defaults
        for key, column in inspect(cls).columns.items():
            if key not in kwargs and column.default is not None:
                arg = column.default.arg
                column_default = arg if callable(arg) else lambda: arg
                setattr(instance, key, column_default(instance))
        return instance

    @classmethod
    def create_and_save(cls, db_session=None, **kwargs):
        m = cls.create(**kwargs)
        with db_api.magic_session(db_session) as db:
            db.add(m)
            db.commit()
            return m

    def as_dict(self):
        d = dict(self.__dict__)
        d.pop('_sa_instance_state', None)
        return d

    def update(self, **kwargs):
        for key in kwargs:
            setattr(self, key, kwargs[key])

    def delete(self, db_session=None):
        db = db_session or inspect(self).session
        db.delete(self)

    created_at = sa.Column(sa.DateTime, default=_get_date, nullable=False)
    updated_at = sa.Column(sa.DateTime, default=_get_date, onupdate=_get_date, nullable=False)


class A10BaseMixin(object):

    id = sa.Column(sa.String(36), primary_key=True, nullable=False, default=_uuid_str)
    tenant_id = sa.Column(sa.String(36), nullable=False)
